@@ -38,4 +38,8 @@
 extern int qddb_errno;
 extern char *qddb_errmsg;
 extern int qddb_stabilizing;
+extern int qddb_maxmem;
+extern int qddb_maxfiles;
+extern char *qddb_default_date_format;
+extern int qddb_datestyle;
 #endif
