@@ -29,6 +29,15 @@
 #define DEFAULT_HASHSIZE	20000
 #define DEFAULT_CACHESIZE	100
 #define HASHLINESIZE		44
+#define BASE10_HASHLINESIZE     HASHLINESIZE
+#define BASE62_HASHLINESIZE     28
+
+#define QDDB_HASHTYPE_DEFAULT	0
+#define QDDB_HASHTYPE_MOD1	1
+#define QDDB_HASHTYPE_MOD2	2
+#define QDDB_HASHTYPE_MOD3	3
+#define QDDB_HASHTYPE_MOD4	4
+#define QDDB_HASHTYPE_MOD5	5
 
 typedef struct {
 		size_t		index, num;
