@@ -33,4 +33,7 @@ typedef struct tcllist {
     char		*value;
 } TCLlist;
 
+void Qddb_ResetTCLLex _ANSI_ARGS_((void));
+void Qddb_ResetTCLParser _ANSI_ARGS_((void));
+
 #endif
