@@ -29,7 +29,6 @@
 
 /* EXPORTED:
  *	Qddb_Convert(Schema *, int, void *, int)
- *	Qddb_FindAttributeNumber(Schema *, int *, int)
  */
 
 /* Entry conversions 
@@ -49,13 +48,14 @@ static void 		*ConvertInternalToXXX _ANSI_ARGS_((Schema *, void *, int));
 
 /* From the parsers...
  */
+void Qddb_ResetEntryParser _ANSI_ARGS_((void));
 void Qddb_EntryParse_Set _ANSI_ARGS_((Schema *));
-InCoreEntry *Qddb_EntryParse_Get();
+InCoreEntry *Qddb_EntryParse_Get _ANSI_ARGS_((void));
 
 void Qddb_TCLParse_Set _ANSI_ARGS_((Schema *));
-InCoreEntry *Qddb_TCLParse_Get();
+InCoreEntry *Qddb_TCLParse_Get _ANSI_ARGS_((void));
 
-void ResetTCLParser _ANSI_ARGS_((void));
+void Qddb_ResetTCLParser _ANSI_ARGS_((void));
 int TCLParse _ANSI_ARGS_((void));
 
 /* Qddb_Convert -- Convert between various QDDB internal structure types.
@@ -82,7 +82,7 @@ void *Qddb_Convert(schema, FromType, FromPtr, ToType)
 	return ConvertInternalToXXX(schema, FromPtr, ToType);
     /* XXX -> Internal */
     retval = ConvertXXXToInternal(schema, FromType, FromPtr);
-#if defined(DIAGNOSTIC)
+#if defined(DEBUG)
     Qddb_SortInCoreEntries(schema, retval);
 #endif
     if (ToType == QDDB_ENTRYTYPE_INTERNAL || retval == NULL)
@@ -94,28 +94,6 @@ void *Qddb_Convert(schema, FromType, FromPtr, ToType)
     return retval2;
 }
 
-int Qddb_FindAttributeNumber(schema, Attributes, HowMany)
-    Schema		*schema;
-    size_t		Attributes[];
-    int			HowMany;
-{
-    size_t		i, j;
-
-    for (i = 1; i <= schema->NumberOfAttributes; i++) {
-	for (j = 1; j < HowMany; j++) {
-	    if (schema->Entries[i].AncestorNumber[j] != Attributes[j-1])
-		break;
-	}
-	if (j == HowMany) {
-	    if (schema->Entries[i].Level != HowMany)
-		continue;
-	    if (Attributes[j-1] == schema->Entries[i].Number)
-		return i;
-	}
-    }
-    return -1;
-}
-
 static void *ConvertInternalToXXX(schema, FromPtr, ToType)
     Schema			*schema;
     void 			*FromPtr;
@@ -187,13 +165,13 @@ static InCoreEntry *ConvertReadableToInternal(schema, input)
     char			*input; /* buffer full of a readable format entry */
 {
     InCoreEntry			*retval;
-    int				ResetEntryParser(), EntryParse();
+    int				EntryParse();
 
     qddb_errno = 0;
     if (qddb_errmsg != NULL)
 	Free(qddb_errmsg);
     qddb_errmsg = NULL;
-    ResetEntryParser();
+    Qddb_ResetEntryParser();
     Qddb_EntryParse_Set(schema);
     Qddb_SetBuffer(input);
     if (EntryParse() != 0)
@@ -206,9 +184,10 @@ static InCoreEntry *ConvertReadableToInternal(schema, input)
 static char *TranslateDataToReadable(data)
     char		*data;
 {
-    char		*ch = Malloc(BUFSIZ);
+    char		*ch;
     int			bufsiz = BUFSIZ, i;
 
+    ch = Malloc(BUFSIZ);
     for (i = 0; *data; i++, data++) {
 	if (bufsiz-3 <= i)
 	    ch = Realloc(ch, (size_t)(bufsiz += BUFSIZ));
@@ -463,6 +442,9 @@ static InCoreEntry *ConvertExternalToInternal(schema, MyEntry)
 	 * add the attribute to the array of InCoreEntries.
 	 */
 	Bitsy = strlen(*MyEntry);
+#if defined(DEBUG)
+	printf("MyEntry: %s\n", *MyEntry);
+#endif
 	if (Bitsy == 0) { /* hack, something is setting this incorrectly */
 	    Free(*MyEntry);
 	    *MyEntry = NULL;
@@ -476,6 +458,9 @@ static InCoreEntry *ConvertExternalToInternal(schema, MyEntry)
 		Itsy++;
 	    (*MyEntry)[Itsy] = '\0';
 	    Attributes[i] = atoi(num);
+#if defined(DEBUG)
+	    printf("Attributes[%d] %d\n", i, (int)Attributes[i]);
+#endif
 	    Itsy++;
     	    num = (*MyEntry)+Itsy;
 	    while ((*MyEntry)[Itsy] != '.' && !isspace((*MyEntry)[Itsy]))
@@ -498,7 +483,7 @@ static InCoreEntry *ConvertExternalToInternal(schema, MyEntry)
             int			j;
 
 	    for (j = 0; j < i; j++)
-		fprintf(stderr, "Attributes[%d] == %d\n", j, Attributes[j]);
+		fprintf(stderr, "Attributes[%d] == %d\n", (int)j, (int)Attributes[j]);
 	    fprintf(stderr, "MyEntry: %s\n", *MyEntry);
 	    PANIC("ConvertExternalToInternal: Qddb_FindAttributeNumber failed");
 	}
@@ -508,7 +493,7 @@ static InCoreEntry *ConvertExternalToInternal(schema, MyEntry)
 	    Data = Qddb_DateTimeToString(*MyEntry+Itsy, 
 					 schema->Entries[RetVal[InRetVal].AttributeNumber].Format);
 #if defined(DEBUG)
-	fprintf(stderr, "Date: %x\n", Data);
+	fprintf(stderr, "Date: %x\n", (unsigned int)Data);
 #endif
 	    RetVal[InRetVal].BytesOfData = strlen(Data);
 	} else {
@@ -518,7 +503,7 @@ static InCoreEntry *ConvertExternalToInternal(schema, MyEntry)
 	    tmp = *MyEntry + Itsy;
 	    Data = Malloc(len+1);
 #if defined(DEBUG)
-	fprintf(stderr, "Data: %x\n", Data);
+	fprintf(stderr, "Data: %x\n", (unsigned int)Data);
 #endif
 	    tmp = *MyEntry + Itsy;
 	    dataspot = Data;
@@ -668,7 +653,7 @@ static InCoreEntry *ConvertTCLvalueToInternal(schema, input)
 #else
     InCoreEntry			*retval;
 
-    ResetTCLParser(); 
+    Qddb_ResetTCLParser(); 
     Qddb_TCLParse_Set(schema);
     Qddb_SetBuffer(input);
     if (TCLParse() != 0)
@@ -706,7 +691,7 @@ static char *ConvertInternalToTCLvalue(schema, ThisEntry)
 	int			flag;
 
 	if (ThisEntry->Data != NULL)
-	    StripBackslashes(ThisEntry->Data);
+	    Qddb_StripBackslashes(ThisEntry->Data);
 	TmpEntry = schema->Entries+ThisEntry->AttributeNumber;
 	if (LastAttr != -1)
 	    LastEntry = schema->Entries+LastAttr;
@@ -890,9 +875,9 @@ static InCoreEntry *ConvertTCLExternalToInternal(schema, entry)
     char			*Data, *AttributeName, *InstanceName;
     int				SequenceNumber = -1;
     InCoreEntry			*incore;
-    int				incore_top = 0, i, j, k;
+    int				incore_top = 0;
     int				first_line = 0;
-    int				attr_num;
+    int				attr_num, level;
 
     if (entry == NULL) {
 	return NULL;
@@ -922,7 +907,6 @@ static InCoreEntry *ConvertTCLExternalToInternal(schema, entry)
 	    entry++;
 	    continue;
 	}
-
 	while (*ch != '\0' && isspace(*ch))
 	    ch++; /* skip leading spaces */
 	AttributeName = ch;
@@ -934,6 +918,7 @@ static InCoreEntry *ConvertTCLExternalToInternal(schema, entry)
 	    ch++;
 	*ch++ = '\0';
 	Data = ch;
+	Qddb_StripBackslashes(Data);
 	incore[incore_top].Data = (char *)Malloc(strlen(Data)+1);
 	if (incore[incore_top].Data == NULL)
 	    return NULL;
@@ -947,20 +932,12 @@ static InCoreEntry *ConvertTCLExternalToInternal(schema, entry)
 	    /* FIXME: free up stuff */
 	    return NULL;
 	}
+	level = schema->Entries[attr_num].Level;
 	incore[incore_top].AttributeNumber = attr_num;
-	i = j = 0;
-	k = 1;
-	if (StringToInstance(InstanceName, incore[incore_top].Instance) == -1) {
+	if (StringToInstance(InstanceName, incore[incore_top].Instance) != level) {
 	    /* FIXME: free up stuff */
 	    return NULL;
 	}
-	while (InstanceName+i < Data) {
-	    while (isdigit(InstanceName[i]))
-		i++;
-	    InstanceName[i++] = '\0';
-	    incore[incore_top].Instance[k++] = atoi(InstanceName+j);
-	    j = i;
-	}
 	incore_top++;
 	entry++;
     }
@@ -1031,13 +1008,14 @@ static char *ConvertInternalToTCLExternal(schema, incore)
 #endif
 }
 
-static void DataTreeAttributes _ANSI_ARGS_((Schema *, InCoreEntry **, DataTree **, int *));
+static void DataTreeAttributes _ANSI_ARGS_((Schema *, InCoreEntry **, DataTree **, int *, int *));
 
-static void DataTreeInstances(schema, entry, tree_inst, instances)
+static void DataTreeInstances(schema, entry, tree_inst, instances, currnum)
     Schema			*schema;
     InCoreEntry			**entry;
     DataTree			*tree_inst;
     int				*instances;
+    int				*currnum;
 {
     SchemaNode			*schema_node;
     int				level, *newinstances, i, j, num;
@@ -1052,21 +1030,26 @@ static void DataTreeInstances(schema, entry, tree_inst, instances)
 	    if (tree_inst[i].datatree_type != DATATREE_CHILDREN)
 		abort();
 	    newinstances[level-1] = i+1;
-	    DataTreeAttributes(schema, entry, tree_inst[i].datatree_children, newinstances);
+	    DataTreeAttributes(schema, entry, tree_inst[i].datatree_children, newinstances, currnum);
 	}
 	return;
     }
-    for (num = 0; (*entry)[num].SequenceNumber != 0; num++);
+    num = *currnum;
     for (i = 0; tree_inst[i].datatree_type != DATATREE_END; i++) {
 	InCoreEntry		*tmp_entry;
 	int			tmp;
 
-	if (tree_inst[i].datatree_type == DATATREE_NOVAL) {
+	if (tree_inst[i].datatree_type == DATATREE_NOVAL || tree_inst[i].modified == False) {
+#if defined(DATATREE_DEBUG)
+	    fprintf(stderr, "skipping\n");
+#endif
 	    continue;
 	}
 	newinstances[level-1] = i+1;
 	num++;
+#if defined(notdef)
 	*entry = (InCoreEntry *)Realloc(*entry, sizeof(InCoreEntry)*(num+1));
+#endif
 	(*entry)[num].SequenceNumber = 0;
 	tmp = num-1;
 	tmp_entry = (*entry) + tmp;
@@ -1074,6 +1057,14 @@ static void DataTreeInstances(schema, entry, tree_inst, instances)
 	fprintf(stderr, "seq: %d\n", tree_inst[i].datatree_sequence_number);
 #endif
 	tmp_entry->SequenceNumber = tree_inst[i].datatree_sequence_number;
+#if defined(DATATREE_DEBUG)
+	if (tmp_entry->SequenceNumber == 0)
+	    PANIC("Bad sequence Number!!!\n");
+#else
+	if (tmp_entry->SequenceNumber == 0) {
+	    tmp_entry->SequenceNumber = -1;
+	}
+#endif
 	tmp_entry->Index = tmp;
 	tmp_entry->Marked = Inapplicable;
 	tmp_entry->Data = Qddb_DataTreeProcess(schema, NULL, tree_inst+i, QDDB_DATATREE_PROC_GETDATA, 0);
@@ -1087,34 +1078,79 @@ static void DataTreeInstances(schema, entry, tree_inst, instances)
 	    tmp_entry->Instance[j] = newinstances[j];
 	}
     }
+    *currnum = num;
 }
 
-static void DataTreeAttributes(schema, entry, tree, instances)
+static void DataTreeAttributes(schema, entry, tree, instances, currnum)
     Schema			*schema;
     InCoreEntry			**entry;
     DataTree			**tree;
     int				*instances;
+    int				*currnum;
 {
     while (*tree != NULL) {
 #if defined(DATATREE_DEBUG)
 	fprintf(stderr, "intermediate node %s\n", (*tree)->datatree_schema->schemanode->Name);
 #endif
-	DataTreeInstances(schema, entry, (*tree), instances);
+	DataTreeInstances(schema, entry, (*tree), instances, currnum);
 	tree++;
     }
 }
 
+static int CountDataTreeLeaves _ANSI_ARGS_((DataTree **));
+
+static int CountDataTreeInstances(tree_inst)
+    DataTree			*tree_inst;
+{
+    int				i, retval = 0;
+    SchemaNode			*schema_node;
+
+    schema_node = tree_inst->datatree_schema->schemanode;
+    if (schema_node->IsLeaf == True) {
+	for (i = 0; tree_inst[i].datatree_type != DATATREE_END; i++);
+	retval = i;
+    } else {
+	for (i = 0; tree_inst[i].datatree_type != DATATREE_END; i++)
+	    retval += CountDataTreeLeaves(tree_inst[i].datatree_children);
+    }
+    return retval;
+}
+
+static int CountDataTreeLeaves(tree)
+    DataTree			**tree;
+{
+    int				retval = 0;
+
+    while (*tree != NULL) {
+	retval += CountDataTreeInstances(*tree);
+	tree++;
+    }
+    return retval;
+}
+
 static InCoreEntry *ConvertDataTreeToInternal(schema, tree)
     Schema			*schema;
     DataTree			**tree;
 {
+    int				number_leaves;
     InCoreEntry			*retval = NULL;
+    int				num = 0;
 
     if (tree == NULL || *tree == NULL)
 	return NULL;
-    retval = (InCoreEntry *)Malloc(sizeof(InCoreEntry));
+    number_leaves = CountDataTreeLeaves(tree) + 1;
+#if defined(DEBUG)
+    fprintf(stderr, "expect to traverse %d attributes\n", number_leaves);
+#endif
+    retval = (InCoreEntry *)Malloc(sizeof(InCoreEntry)*number_leaves);
     retval->SequenceNumber = 0;
-    DataTreeAttributes(schema, &retval, tree, NULL);
+    DataTreeAttributes(schema, &retval, tree, NULL, &num);
+#if defined(DEBUG)
+    fprintf(stderr, "traversed %d attributes\n", num);
+
+    for (num = 0; retval[num].SequenceNumber != 0; num++);
+    fprintf(stderr, "counted %d attributes\n", num);
+#endif
     return retval;
 }
 
