@@ -28,4 +28,9 @@
 
 int qddb_errno = 0;
 char *qddb_errmsg = NULL;
-int qddb_stabilizing = 0;
+int qddb_stabilizing = False;
+int qddb_maxmem = MAXIMUM_MEMORY_USAGE;
+int qddb_maxfiles = 128;
+char *qddb_default_date_format = DEFAULT_DATE_FORMAT;
+int qddb_datestyle;
+
