@@ -1,7 +1,7 @@
 
 /* qddb/Lib/LibQddb/Index.c
  *
- * Copyright (C) 1993, 1994, 1995 Herrin Software Development, Inc.
+ * Copyright (C) 1993-1997 Herrin Software Development, Inc.
  * All rights reserved.
  *
  * This file is part of Qddb.
@@ -27,7 +27,6 @@
 #include "Qddb.h"
 
 /* EXPORTED:
- * 	int Qddb_HashValue(char *)
  *	void Qddb_MakeIndex(char *)
  */
 
@@ -45,40 +44,32 @@ static KeyNode *HashEntryToKeyNodes _ANSI_ARGS_((Schema *, FILE *, Hash *));
 static void PurgeHashList _ANSI_ARGS_((Bucket *, size_t));
 static void PurgeHashBucket _ANSI_ARGS_((KeyNode *));
 static void PurgePosNode _ANSI_ARGS_((PosNode *));
-static KeyNode *IndexParse _ANSI_ARGS_((Schema *, Entry, int *));
 static void AddToHashList _ANSI_ARGS_((Bucket *, KeyNode *, off_t, size_t));
 static void WriteHashList _ANSI_ARGS_((Schema *, FILE *, FILE *, Bucket *, Boolean, size_t, size_t));
-static void WriteKeyIndex _ANSI_ARGS_((FILE *, KeyIndexIn *));
+static void WriteKeyIndex _ANSI_ARGS_((Schema *, FILE *, KeyIndexIn *));
 static void SortKeyIndex _ANSI_ARGS_((size_t, size_t));
-static void WriteNumericIndex _ANSI_ARGS_((FILE *, NumericIndexIn *));
+static void WriteNumericIndex _ANSI_ARGS_((Schema *, FILE *, NumericIndexIn *));
 static void SortNumericIndex _ANSI_ARGS_((size_t, size_t));
+static void PurgeKeyIndex _ANSI_ARGS_((Schema *, KeyIndexIn *));
+static void PurgeNumericIndex _ANSI_ARGS_((Schema *, NumericIndexIn *));
 
-int Qddb_HashValue(schema, string)
-    Schema		*schema;
-    char		*string;
+static int  ptr_strcmp1 _ANSI_ARGS_((char *, char **));
+
+static int ptr_strcmp1(str1, str2)
+    char                *str1, **str2;
 {
-    int		RetVal, len;
-
-    len = strlen(string);
-    if (len > 0)
-	RetVal = (int)string[0]*17576;
-    else
-	return 0;
-    if (len > 1)
-    	RetVal += (int)string[1]*676;
-    else
-    	return RetVal % schema->HashSize;
-    if (len > 2)
-    	RetVal += (int)string[2]*26;
-    else 
-    	return RetVal % schema->HashSize;
-    if (len > 3)
-    	RetVal += (int)string[3];
-    return RetVal % schema->HashSize;
+    return strcmp(str1, *str2);
 }
 
-void Qddb_MakeIndex(schema)
+static int ptr_strcmp2(str1, str2)
+    char                **str1, **str2;
+{
+    return strcmp(*str1, *str2);
+}
+
+void Qddb_MakeIndex(schema, verbose)
     Schema		*schema;
+    Boolean             verbose;
 {
     char		RelationFN[MAXFILENAMELEN];
     char		IndexFN[MAXFILENAMELEN];
@@ -86,21 +77,22 @@ void Qddb_MakeIndex(schema)
     int			IndexFile, DatabaseFile, KeyFile;
     FILE		*KeyFilePtr, *IndexFilePtr, *fdopen();
     FILE		**IndexFilePtrs;
-#ifdef SQL
     char		KeyIndexFN[MAXFILENAMELEN], NumericIndexFN[MAXFILENAMELEN];
     int			KeyIndexFile, NumericIndexFile;
     FILE		*KeyIndexFilePtr, *NumericIndexFilePtr;
-#endif
     FILE		**HashFilePtrs;
     FILE		*HashFilePtr;
     Bucket		*HashList;
     Entry		ThisEntry = NULL;
-    int			Start, Length;
+    Entry               ExcludeEntry = NULL;
+    off_t		Start;
+    size_t              Length;
+    char                start_ch[BUFSIZ], length_ch[BUFSIZ];
     int			i, j, Error;
+    size_t              nmemb = 0;
     KeyNode		*List;
-#if defined(MAXIMUM_MEMORY_USAGE)
+    int                 entrysize, hashlinesize, base;
     int			idxnum = 0;
-#endif
     size_t		LastHash;
 #if defined(USE_MALLOC_STATS)
 #if defined(QDDB_STAB_STATS)
@@ -112,11 +104,9 @@ void Qddb_MakeIndex(schema)
     size_t		starting_bytes;
 #endif
 
-#ifdef SQL
     NumberOfKeys = 0;
     NumberOfNumericKeys = 0;
     NumericIndexTop = 0;
-#endif
     KeyIndex = NULL;
     KeyIndexTop = 0;
     strcpy(RelationFN, schema->RelationName);
@@ -126,20 +116,29 @@ void Qddb_MakeIndex(schema)
     strcat(HashFN, "/HashTable");
     strcpy(KeyFN, RelationFN);
     strcat(KeyFN, "/Database.key");
-#ifdef SQL
     strcpy(KeyIndexFN, RelationFN);
     strcat(KeyIndexFN, "/KeyIndex");
     strcpy(NumericIndexFN, RelationFN);
     strcat(NumericIndexFN, "/NumericIndex");
-#endif
     IndexFile = Open(IndexFN, O_WRONLY|O_CREAT|O_TRUNC, 0666);
     IndexFilePtr = fdopen(IndexFile, "w");
-#ifdef SQL
     KeyIndexFile = Open(KeyIndexFN, O_RDWR|O_CREAT|O_TRUNC, 0666);
     KeyIndexFilePtr = fdopen(KeyIndexFile, "w");
     NumericIndexFile = Open(NumericIndexFN, O_RDWR|O_CREAT|O_TRUNC, 0666);
     NumericIndexFilePtr = fdopen(NumericIndexFile, "w");
-#endif
+    if (schema->UseCondensedIndexing == False) {
+	base = 10;
+	hashlinesize = BASE10_HASHLINESIZE;
+    } else {
+	base = 62;
+	hashlinesize = BASE62_HASHLINESIZE;
+    }
+    entrysize = hashlinesize / 4;
+    if (schema->UseExcludeWords == True) {
+	ExcludeEntry = Qddb_ReadExcludeWords(schema, &nmemb);
+	qsort(ExcludeEntry, nmemb, sizeof(char *),
+	      (int (*) _ANSI_ARGS_((const void *, const void *)))ptr_strcmp2);
+    }
     KeyFile   = Open(KeyFN, O_RDONLY, 0);
     KeyFilePtr = fdopen(KeyFile, "r");
     DatabaseFile = schema->database_fd;
@@ -164,9 +163,11 @@ void Qddb_MakeIndex(schema)
     }
     for (i = 0; i < schema->HashSize; i++)
     	HashList[i] = NULL;
-    while (fscanf(KeyFilePtr, "%d %d\n", &Start, &Length) != EOF) {
+    while (fscanf(KeyFilePtr, "%s %s\n", start_ch, length_ch) != EOF) {
+	Start = Qddb_BaseToOffset(start_ch, base);
+	Length = Qddb_BaseToSize(length_ch, base);
 #if defined(USE_MALLOC_STATS)
-	if (++recnum % 100 == 0) {
+	if ((++recnum % 100) == 0 && verbose) {
 #if defined(QDDB_STAB_STATS)
 	    printf("%d Records Processed\n", recnum);
 #endif
@@ -180,9 +181,9 @@ chunks_free(%u) bytes_free(%u)\n",
 #if defined(MAXIMUM_MEMORY_USAGE)
 #if defined(USE_GNU_MALLOC)
 	m = mstats();
-	if (m.bytes_used > (size_t)(starting_bytes+MAXIMUM_MEMORY_USAGE*1024)) {
+	if (m.bytes_used > (size_t)(starting_bytes+qddb_maxmem*1024)) {
 #else
-	if (qddb_malloc_bytes_used > (size_t)(MAXIMUM_MEMORY_USAGE*1024)) {
+	if (qddb_malloc_bytes_used > (size_t)(qddb_maxmem*1024)) {
 #endif
 	    FILE		*NewIndexFilePtr, *NewHashFilePtr;
 	    char		buf[BUFSIZ];
@@ -204,53 +205,64 @@ chunks_free(%u) bytes_free(%u)\n",
 	    fclose(NewHashFilePtr);
 #if defined(DIAGNOSTIC) && defined(USE_GNU_MALLOC) && defined(USE_MALLOC_STATS)
 	    m = mstats();
-	    if (m.bytes_used != starting_bytes)
+	    if (m.bytes_used != starting_bytes && verbose)
 		fprintf(stdout, "After purge, starting bytes (%d) != bytes used (%d)\n",
 			(int)starting_bytes, (int)m.bytes_used);
 #endif
 	}
 #endif
 	Qddb_ReadEntry(DatabaseFile, &ThisEntry, (off_t)Start, (size_t)Length, False);
-    	List = IndexParse(schema, ThisEntry+1, &Error);
+    	List = Qddb_IndexParse(schema, ThisEntry+1, &Error);
 	if (Error) {
 	    fprintf(stdout, "Cannot index the following tuple:\n%s\n", ThisEntry[0]);
 	    fflush(stdout);
-	    exit(1);
+	    List = NULL;
 	}
     	while (List != NULL) {
 	    int		Value;
 	    KeyNode	*next;
 
+	    if (schema->UseExcludeWords == True && nmemb > 0) {
+		if (bsearch(List->s, ExcludeEntry, nmemb, sizeof(char *), 
+			    (int (*) _ANSI_ARGS_((const void *, const void *)))ptr_strcmp1) != NULL) {
+		    next = List->next;
+		    if (List->inst != NULL)
+			Free(List->inst);
+		    Free(List->s);
+		    Free(List);
+		    List = next;
+		    continue;
+		}
+	    }
 	    Value = Qddb_HashValue(schema, List->s);
 	    next = List->next;
 	    AddToHashList(&HashList[Value], List, (off_t)Start, (size_t)Length); /* Frees up List */
 	    List = next;
     	}
+#if !defined(USE_GNU_MALLOC)
+	qddb_malloc_bytes_used -= Length*2; /* bad estimate, but it works mostly ok. */
+#endif
     }
     if (ThisEntry != NULL)
 	Qddb_Free(QDDB_TYPE_ENTRY, ThisEntry);
-#ifdef SQL
-#if defined(DIAGNOSTIC)
-    fprintf(stdout, "Allocating %d keys\n", (int)NumberOfKeys);
+    if (ExcludeEntry != NULL)
+	Qddb_Free(QDDB_TYPE_ENTRY, ExcludeEntry);
+    if (verbose)
+	fprintf(stdout, "Allocating %d alphanumeric keys\n", (int)NumberOfKeys);
     if (NumberOfKeys == 0)
 	return;
-#endif
     KeyIndex = (KeyIndexIn *)Malloc(sizeof(KeyIndexIn)*NumberOfKeys);
-    /* This wastes some space, but the computation necessary to correct it
-     *	is large.  Trade space for time.
-     */
-    NumericIndex = (NumericIndexIn *)Malloc(sizeof(NumericIndexIn)*NumberOfKeys);
-#endif
-#if defined(MAXIMUM_MEMORY_USAGE)
+    if (verbose)
+	fprintf(stdout, "Allocating %d numeric keys\n", (int)NumberOfNumericKeys);
+    NumericIndex = (NumericIndexIn *)Malloc(sizeof(NumericIndexIn)*NumberOfNumericKeys);
     if (idxnum == 0) {
-#endif
 	WriteHashList(schema, IndexFilePtr, HashFilePtr, HashList, 
 		      True, (size_t)0, (size_t)schema->HashSize);
-#ifdef SQL
-	WriteKeyIndex(KeyIndexFilePtr, KeyIndex);
-	WriteNumericIndex(NumericIndexFilePtr, NumericIndex);
-#endif
-#if defined(MAXIMUM_MEMORY_USAGE)
+	PurgeHashList(HashList, schema->HashSize);
+	WriteKeyIndex(schema, KeyIndexFilePtr, KeyIndex);
+	PurgeKeyIndex(schema, KeyIndex);
+	WriteNumericIndex(schema, NumericIndexFilePtr, NumericIndex);
+	PurgeNumericIndex(schema, NumericIndex);
     } else {
 	FILE		*NewIndexFilePtr, *NewHashFilePtr;
 	char		buf[BUFSIZ];
@@ -310,7 +322,7 @@ chunks_free(%u) bytes_free(%u)\n",
 	    for (j = 0; j < idxnum; j++) { /* each (HashTable-*,Index-*) */
 		Hash			hash;
 		KeyNode			*key;
-		char			*ch, *ch1, buf[BUFSIZ];
+		char			buf[BASE10_HASHLINESIZE+1];
 
 		/* 1. Read HashEntry
 		 * 2. Read Index pointed to by HashEntry
@@ -318,10 +330,15 @@ chunks_free(%u) bytes_free(%u)\n",
 		 * 4. Add them to HashList
 		 */
 		fgets(buf, sizeof(buf), HashFilePtrs[j]);
-		hash.num = (size_t)strtoul(buf, &ch, 10);
-		hash.index = (size_t)strtoul(ch, &ch1, 10);
-		hash.pos = (off_t)strtoul(ch1, &ch, 10);
-		hash.len = (size_t)strtoul(ch, &ch1, 10);
+		buf[hashlinesize] = '\0';
+		buf[entrysize-1]  = '\0';
+		buf[(entrysize*2)-1]  = '\0';
+		buf[(entrysize*3)-1]  = '\0';
+		buf[(entrysize*4)-1]  = '\0';
+		hash.num = Qddb_BaseToSize(buf, base);
+		hash.index = Qddb_BaseToSize(buf+entrysize, base);
+		hash.pos = Qddb_BaseToOffset(buf+(2*entrysize), base);
+		hash.len = Qddb_BaseToSize(buf+(3*entrysize), base);
 		key = HashEntryToKeyNodes(schema, IndexFilePtrs[j], &hash);
 		while (key != NULL) {
 		    KeyNode	*next;
@@ -342,16 +359,16 @@ chunks_free(%u) bytes_free(%u)\n",
 #if defined(USE_GNU_MALLOC)
 	    m = mstats();
 #if defined(USE_MALLOC_STATS)
-	    if (i % 100 == 0) {
+	    if (i % 100 == 0 && verbose) {
 		printf("Hash #%d: bytes_total(%u) chunks_used(%u) bytes_used(%u) \
 chunks_free(%u) bytes_free(%u)\n", i,
 		   m.bytes_total, m.chunks_used, m.bytes_used, m.chunks_free, m.bytes_free);
 		fflush(stdout);
 	    }
 #endif
-	    if (m.bytes_used > (size_t)(starting_bytes+MAXIMUM_MEMORY_USAGE*1024)) {
+	    if (m.bytes_used > (size_t)(starting_bytes+qddb_maxmem*1024)) {
 #else
-	    if (qddb_malloc_bytes_used > (size_t)(MAXIMUM_MEMORY_USAGE*1024)) {
+	    if (qddb_malloc_bytes_used > (size_t)(qddb_maxmem*1024)) {
 #endif
 		WriteHashList(schema, IndexFilePtr, HashFilePtr, HashList, 
 			      True, LastHash, (size_t)i+1);
@@ -381,12 +398,32 @@ chunks_free(%u) bytes_free(%u)\n", i,
 	WriteHashList(schema, IndexFilePtr, HashFilePtr, HashList, 
 		      True, LastHash, (size_t)schema->HashSize);
 	PurgeHashList(HashList, schema->HashSize);
-#ifdef SQL
-	WriteKeyIndex(KeyIndexFilePtr, KeyIndex);
-	WriteNumericIndex(NumericIndexFilePtr, NumericIndex);
-#endif
+	WriteKeyIndex(schema, KeyIndexFilePtr, KeyIndex);
+	PurgeKeyIndex(schema, KeyIndex);
+	WriteNumericIndex(schema, NumericIndexFilePtr, NumericIndex);
+	PurgeNumericIndex(schema, NumericIndex);
     }
-#endif /* MAXIMUM_MEMORY_USAGE */
+    Free(HashList);
+}
+
+static void PurgeKeyIndex(schema, KeyIndex)
+    Schema              *schema;
+    KeyIndexIn		*KeyIndex;
+{
+    int                 i;
+
+    for (i = 0; i < KeyIndexTop; i++) {
+	if (KeyIndex[i].s != NULL)
+	    Free(KeyIndex[i].s);
+    }
+    Free(KeyIndex);
+}
+
+static void PurgeNumericIndex(schema, NumericIndex)
+    Schema              *schema;
+    NumericIndexIn	*NumericIndex;
+{
+    Free(NumericIndex);
 }
 
 static void PurgePosNode(List)
@@ -413,10 +450,9 @@ static void PurgeHashBucket(List)
 	PurgePosNode(List->pos);
 	if (List->s != NULL)
 	    Free(List->s);
-#if defined(notdef)
+/* notdef */
 	if (List->inst != NULL)
 	    Free(List->inst);
-#endif
 	Free(List);
 	List = newlist;
     }
@@ -434,7 +470,7 @@ static void PurgeHashList(List, Size)
     }
 }
 
-static KeyNode *IndexParse(schema, ThisEntry, ErrorPtr)
+KeyNode *Qddb_IndexParse(schema, ThisEntry, ErrorPtr)
     Schema		*schema;
     Entry		ThisEntry;
     int			*ErrorPtr;
@@ -463,7 +499,7 @@ static KeyNode *IndexParse(schema, ThisEntry, ErrorPtr)
 	    strcat(full_attr, schema->ReducedAttributeTable[attr]);
 	    if (Qddb_TranslateExternalID(schema, full_attr, &attr, &tmp) != 0) {
 		fprintf(stderr, "Warning: attribute %s malformed, skipping attr %d..\n", 
-			schema->ReducedAttributeTable[attr], attr);
+			schema->ReducedAttributeTable[attr], (int)attr);
 		Free(inst);
 		Free(full_attr);
 		ThisEntry++;
@@ -496,15 +532,12 @@ static KeyNode *IndexParse(schema, ThisEntry, ErrorPtr)
 	    List->next = ptr;
 	    List->s = Malloc((size_t)len+1);
 	    strcpy(List->s, token);
+	    Qddb_StripBackslashes(List->s);
 	    List->pos = NULL;
 	    List->len = len;
-#ifdef SQL
 	    List->attr = attr;
-#ifdef INSTANCE_QUALIFICATION
 	    List->inst = (char *)Malloc(strlen(inst)+1);
 	    strcpy(List->inst, inst);
-#endif
-#endif
 	}
 	ThisEntry++;
 	Free(inst);
@@ -521,9 +554,9 @@ static void AddToHashList(List, Node, Start, Length)
     KeyNode		*ptr;
 
     if (*List == NULL) {
-#ifdef SQL
 	NumberOfKeys++;
-#endif
+	if (Qddb_IsNumericString(Node->s))
+	    NumberOfNumericKeys++;
 	*List = (KeyNode *)Malloc(sizeof(KeyNode));
 	**List = *Node;
 	(*List)->inst = NULL;
@@ -532,12 +565,8 @@ static void AddToHashList(List, Node, Start, Length)
 	(*List)->pos->Start = Start;
 	(*List)->pos->Length = Length;
 	(*List)->pos->next = NULL;
-#ifdef SQL
 	(*List)->pos->attr = Node->attr;
-#ifdef INSTANCE_QUALIFICATION
 	(*List)->pos->inst = Node->inst;
-#endif
-#endif
 	Free(Node);
 	return;
     }
@@ -549,12 +578,8 @@ static void AddToHashList(List, Node, Start, Length)
 	    PosPtr = (PosNode *)Malloc(sizeof(PosNode));
 	    PosPtr->Start = Start;
 	    PosPtr->Length = Length;
-#ifdef SQL
 	    PosPtr->attr = Node->attr;
-#ifdef INSTANCE_QUALIFICATION
 	    PosPtr->inst = Node->inst;
-#endif
-#endif
 	    PosPtr->next = ptr->pos;
 	    ptr->pos = PosPtr;
 	    /* Free up the word and node. */
@@ -564,9 +589,9 @@ static void AddToHashList(List, Node, Start, Length)
 	}
 	ptr = ptr->next;
     }
-#ifdef SQL
     NumberOfKeys++;
-#endif
+    if (Qddb_IsNumericString(Node->s))
+	NumberOfNumericKeys++;
     ptr = *List;
     *List = (KeyNode *)Malloc(sizeof(KeyNode));
     **List = *Node;
@@ -575,12 +600,8 @@ static void AddToHashList(List, Node, Start, Length)
     (*List)->pos  = (PosNode *)Malloc(sizeof(PosNode));
     (*List)->pos->Start = Start;
     (*List)->pos->Length = Length;
-#ifdef SQL
     (*List)->pos->attr = Node->attr;
-#ifdef INSTANCE_QUALIFICATION
     (*List)->pos->inst = Node->inst;
-#endif
-#endif
     (*List)->pos->next = NULL;
     Free(Node);
 }
@@ -594,16 +615,28 @@ static void WriteHashList(schema, IndexFilePtr, HashFilePtr, HashList, PassTwo,
     size_t		Begin, End;
 {
     KeyNode		*Key;
-    int			Start, Length;
-    int			i, NumberHashed; 
+    off_t		Start;
+    size_t              Length;
+    size_t		i, NumberHashed; 
+    int			file_number;
     Boolean		IsALine = False;
-#ifdef SQL
+    int                 base, entrysize, hashlinesize;
+    char                num[BASE10_HASHLINESIZE], index[BASE10_HASHLINESIZE];
+    char                len[BASE10_HASHLINESIZE], pos[BASE10_HASHLINESIZE];
     off_t		WhereAmI, WhereAmINow;
-#endif
     
+    if (schema->UseCondensedIndexing == False) {
+	base = 10;
+	hashlinesize = BASE10_HASHLINESIZE;
+    } else {
+	base = 62;
+	hashlinesize = BASE62_HASHLINESIZE;
+    }
+    entrysize = (hashlinesize / 4) - 1;
     for (i = Begin; i < End; i++) {
 	Key = HashList[i];
-	Start = lseek(fileno(IndexFilePtr), (off_t)0, 1);
+	file_number = (int)((unsigned int)fileno(IndexFilePtr));
+	Start = lseek(file_number, (off_t)0, 1);
 	if (Start == -1)
 	    PANIC("Cannot seek to Start");
 	NumberHashed = 0;
@@ -613,7 +646,6 @@ static void WriteHashList(schema, IndexFilePtr, HashFilePtr, HashList, PassTwo,
 	    
 	    NumberHashed++;
 	    IsALine = True;
-#ifdef SQL
 	    /* Caution: This fflush(3) is needed for the lseek(2) to
 	     *	find where we are now.
 	     */
@@ -622,10 +654,10 @@ static void WriteHashList(schema, IndexFilePtr, HashFilePtr, HashList, PassTwo,
 	     * 	Need to keep each key and its locator (start,length)
 	     *	to the Index file.
 	     */
-	    WhereAmI = lseek(fileno(IndexFilePtr), (off_t)0, 1);
+	    file_number = (int)((unsigned int)fileno(IndexFilePtr));
+	    WhereAmI = lseek(file_number, (off_t)0, 1);
 	    if (Start == -1)
 		PANIC("SQL: lseek failed");
-#endif
 	    outchar = Key->s;
 	    while (*outchar != '\0') {
 		if (*outchar == ' ') {
@@ -642,30 +674,37 @@ static void WriteHashList(schema, IndexFilePtr, HashFilePtr, HashList, PassTwo,
 	    putc(' ', IndexFilePtr);
 	    ptr = Key->pos;
 	    while (ptr != NULL) {
-#ifdef SQL
-#ifdef INSTANCE_QUALIFICATION
+		char                  *start, *length;
+
 		if (schema->UseReducedAttributeIdentifiers == True) {
-		    fprintf(IndexFilePtr, "%d %d %s ", (int)ptr->Start,
-			    (int)ptr->Length, ptr->inst);
+		    if (schema->UseCondensedIndexing == False) {
+			start = Qddb_OffsetToBase(ptr->Start, 10);
+			length = Qddb_SizeToBase(ptr->Length, 10);
+			fprintf(IndexFilePtr, "%s %s %s ", start, length, ptr->inst);
+		    } else {
+			start = Qddb_OffsetToBase(ptr->Start, 62);
+			length = Qddb_SizeToBase(ptr->Length, 62);
+			fprintf(IndexFilePtr, "%s %s %s ", start, length, ptr->inst);
+		    }
 		} else {
-		    fprintf(IndexFilePtr, "%d %d %d %s ", (int)ptr->Start,
-			    (int)ptr->Length, (int)ptr->attr, ptr->inst);
+		    if (schema->UseCondensedIndexing == False) {
+			start = Qddb_OffsetToBase(ptr->Start, 10);
+			length = Qddb_SizeToBase(ptr->Length, 10);
+			fprintf(IndexFilePtr, "%s %s %d %s ", start, length, (int)ptr->attr, ptr->inst);
+		    } else {
+			start = Qddb_OffsetToBase(ptr->Start, 62);
+			length = Qddb_SizeToBase(ptr->Length, 62);
+			fprintf(IndexFilePtr, "%s %s %d %s ", start, length, (int)ptr->attr, ptr->inst);
+		    }
 		}
-#else
-		fprintf(IndexFilePtr, "%d %d %d ", (int)ptr->Start,
-			(int)ptr->Length, (int)ptr->attr);
-#endif
-#else
- 		fprintf(IndexFilePtr, "%d %d ", (int)ptr->Start, (int)ptr->Length);
-#endif
 		ptr = ptr->next;
 	    }
-#ifdef SQL
 	    /* Caution: This fflush(3) is needed for the lseek(2) to
 	     *	find where we are now.
 	     */
 	    fflush(IndexFilePtr);
-	    WhereAmINow = lseek(fileno(IndexFilePtr), (off_t)0, 1);
+	    file_number = (int)((unsigned int)fileno(IndexFilePtr));
+	    WhereAmINow = lseek(file_number, (off_t)0, 1);
 	    if (PassTwo == True) {
 		/* For KeyIndex:
 		 *	Need the length now.... WhereAmINow-WhereAmI is a good
@@ -680,43 +719,50 @@ static void WriteHashList(schema, IndexFilePtr, HashFilePtr, HashList, PassTwo,
 		 *	If the key is numeric, convert to floating point and
 		 *	save it away for NumericIndex.
 		 */
-		if (IsNumericString(Key->s) == True) {
-		    NumericIndex[NumericIndexTop].s = strtod(Key->s, NULL);
+		if (Qddb_IsNumericString(Key->s) == True) {
+		    NumericIndex[NumericIndexTop].s = Qddb_GetNum(Key->s);
 		    NumericIndex[NumericIndexTop].Start = WhereAmI;
 		    NumericIndex[NumericIndexTop].Length = WhereAmINow - WhereAmI;
 		    NumericIndexTop++;
 		}
 	    }
-#endif
 	    fprintf(IndexFilePtr, "\n");
 	    fflush(IndexFilePtr);
 	    Key = Key->next;
 	}
 	if (IsALine == True) {
-	    Length = lseek(fileno(IndexFilePtr), (off_t)0, 1); 
+	    file_number = (int)((unsigned int)fileno(IndexFilePtr));
+	    Length = lseek(file_number, (off_t)0, 1); 
 	    if (Length == -1)
 		PANIC("Cannot seek to Length");
 	    Length -= Start;
 	    if (Length != 0) {
-		if (PassTwo == False || schema->UseCachedHashing == True)
-		    fprintf(HashFilePtr, "%10d %10d %10d %10d\n", 
-			    NumberHashed, i, Start, Length);
-		else
-		    fprintf(HashFilePtr, "%d %d %d %d\n", 
-			    NumberHashed, i, Start, Length);
+		strcpy(num, Qddb_SizeToBase(NumberHashed, base));
+		strcpy(index, Qddb_SizeToBase(i, base));
+		strcpy(pos, Qddb_OffsetToBase(Start, base));
+		strcpy(len, Qddb_SizeToBase(Length, base));
+		if (PassTwo == False || schema->UseCachedHashing == True) {
+		    fprintf(HashFilePtr, "%*s %*s %*s %*s\n", entrysize, num, entrysize, index, 
+			    entrysize, pos, entrysize, len);
+		} else {
+		    fprintf(HashFilePtr, "%s %s %s %s\n", num, index, pos, len);
+		}
 	    }
 	    IsALine = False;
-	} else if (PassTwo == False || schema->UseCachedHashing == True)
-	    fprintf(HashFilePtr, "%10d %10d %10d %10d\n", 0, i, 0, 0);
+	} else if (PassTwo == False || schema->UseCachedHashing == True) {
+	    fprintf(HashFilePtr, "%*d %*s %*d %*d\n", entrysize, 0, 
+		    entrysize, Qddb_SizeToBase(i, base), entrysize, 0, entrysize, 0);
+	}
     }
 }
 
-static void WriteKeyIndex(KeyIndexFilePtr, KeyIndex)
+static void WriteKeyIndex(schema, KeyIndexFilePtr, KeyIndex)
+    Schema              *schema;
     FILE		*KeyIndexFilePtr;
     KeyIndexIn		*KeyIndex;
 {
     int			i;
-    char		*outchar;
+    char		*outchar, *start, *length;
 
     /* Sort the array by name (method: recursive quicksort).
      */
@@ -738,7 +784,14 @@ static void WriteKeyIndex(KeyIndexFilePtr, KeyIndex)
 	    outchar++;
 	}
 	putc(' ', KeyIndexFilePtr);
-	fprintf(KeyIndexFilePtr, "%d %d\n", (int)KeyIndex[i].Start, (int)KeyIndex[i].Length);
+	if (schema->UseCondensedIndexing == False) {
+	    start = Qddb_OffsetToBase(KeyIndex[i].Start, 10);
+	    length = Qddb_SizeToBase(KeyIndex[i].Length, 10);
+	} else {
+	    start = Qddb_OffsetToBase(KeyIndex[i].Start, 62);
+	    length = Qddb_SizeToBase(KeyIndex[i].Length, 62);
+	}
+	fprintf(KeyIndexFilePtr, "%s %s\n", start, length);
     }
 }
 
@@ -769,11 +822,13 @@ static void SortKeyIndex(l, r)
 }
 
 
-static void WriteNumericIndex(NumericIndexFilePtr, NumericIndex)
+static void WriteNumericIndex(schema, NumericIndexFilePtr, NumericIndex)
+    Schema              *schema;
     FILE		*NumericIndexFilePtr;
     NumericIndexIn	*NumericIndex;
 {
     int			i;
+    char                *start, *length;
     void		SortNumericIndex();
 
     /* Sort the array by numeric key (method: recursive quicksort).
@@ -781,9 +836,16 @@ static void WriteNumericIndex(NumericIndexFilePtr, NumericIndex)
     if (NumericIndexTop > 1)
 	SortNumericIndex(1, NumericIndexTop);
     fprintf(NumericIndexFilePtr, "\n");
-    for (i = 0; i < NumericIndexTop; i++)
-	fprintf(NumericIndexFilePtr, "%f %d %d\n", NumericIndex[i].s, 
-		(int)NumericIndex[i].Start, (int)NumericIndex[i].Length);
+    for (i = 0; i < NumericIndexTop; i++) {
+	if (schema->UseCondensedIndexing == False) {
+	    start = Qddb_OffsetToBase(NumericIndex[i].Start, 10);
+	    length = Qddb_SizeToBase(NumericIndex[i].Length, 10);
+	} else {
+	    start = Qddb_OffsetToBase(NumericIndex[i].Start, 62);
+	    length = Qddb_SizeToBase(NumericIndex[i].Length, 62);
+	}
+	fprintf(NumericIndexFilePtr, "%.17g %s %s\n", NumericIndex[i].s, start, length);
+    }
 }
 
 static void SortNumericIndex(l, r)
@@ -848,7 +910,7 @@ static KeyNode *HashEntryToKeyNodes(schema, IndexFilePtr, HashPtr)
     Hash		*HashPtr;
 {
     char		**KeyEntry, *string, *ch, *Buffer;
-    int			i, EntryNumber, KeyFile = fileno(IndexFilePtr);
+    int			i, EntryNumber, KeyFile = (int)((unsigned int)fileno(IndexFilePtr));
     size_t		NumberRead;
     KeyList		*keylist;
     KeyNode		*keynode, *retval = NULL, *tmp;
