@@ -1,7 +1,7 @@
 
 /* qddb/Lib/LibQddb/KeyList.c 
  *
- * Copyright (C) 1993, 1994 Herrin Software Development, Inc.
+ * Copyright (C) 1993, 1994, 1995 Herrin Software Development, Inc.
  * All rights reserved.
  *
  * This file is part of Qddb.
@@ -35,6 +35,9 @@
 static KeyList *KeyListIntersection _ANSI_ARGS_((Schema *, KeyList *, KeyList *, int));
 static KeyList *KeyListUnion _ANSI_ARGS_((Schema *, KeyList *, KeyList *, int));
 static KeyList *KeyListExclusion _ANSI_ARGS_((Schema *, KeyList *, KeyList *, int));
+static KeyList *KeyListConcat _ANSI_ARGS_((Schema *, KeyList *, KeyList *, int));
+
+static void KeyListMarkAsAny _ANSI_ARGS_((KeyList *));
 
 static KeyList *KeyListMarkEntry _ANSI_ARGS_((Schema *, KeyList *, InCoreEntry *));
 static KeyList *KeyListPruneByRow _ANSI_ARGS_((Schema *, KeyList *, char **, int));
@@ -42,16 +45,25 @@ static KeyList **KeyListSplitList _ANSI_ARGS_((Schema *, KeyList *));
 static KeyList *KeyListGetList _ANSI_ARGS_((KeyList **, int));
 
 static KeyList *PruneListByAttribute _ANSI_ARGS_((Schema *, KeyList *, char *));
+static KeyList *PruneListByIntegerAttribute _ANSI_ARGS_((Schema *, KeyList *, int));
 
 static KeyList *KeyListHashTable _ANSI_ARGS_((KeyList **));
+static void KeyListFreeHashTable _ANSI_ARGS_((KeyList **));
+static KeyList **KeyListBuildHashTable _ANSI_ARGS_((KeyList **, int));
+static KeyList **KeyListHashTableUnion _ANSI_ARGS_((KeyList **, KeyList **, int));
+static KeyList **KeyListHashTableIntersection _ANSI_ARGS_((KeyList **, KeyList **, int));
+static KeyList **KeyListHashTableExclusion _ANSI_ARGS_((KeyList **, KeyList **, int));
+
+static int CompareKeyNodesByRow _ANSI_ARGS_((Schema *, KeyList *, KeyList *));
 
 /* Qddb_KeyListOp -- Standard binary operations on pairs of KeyLists.
  *
  * Operations
  * ----------
- * 	QDDB_KEYLIST_OP_UNION: 			List1 + List2
+ * 	QDDB_KEYLIST_OP_UNION: 			List1 + List2 (state kept)
  * 	QDDB_KEYLIST_OP_INTERSECTION:		List1 * List2
  * 	QDDB_KEYLIST_OP_EXCLUSION:		List1 * ^List2
+ * 	QDDB_KEYLIST_OP_CONCAT:			List1 + List2
  *
  * Flags
  * -----
@@ -81,6 +93,8 @@ KeyList *Qddb_KeyListOp(schema, List1, List2, op, flags)
     case QDDB_KEYLIST_OP_EXCLUSION:
 	retval = KeyListExclusion(schema, List1, List2, flags);
 	break;
+    case QDDB_KEYLIST_OP_CONCAT:
+        retval = KeyListConcat(schema, List1, List2, flags);
     default:
 	qddb_errno = QDDB_ERRNO_INVALID_ARG;
 	return NULL;
@@ -116,6 +130,7 @@ KeyList *Qddb_KeyListOp(schema, List1, List2, op, flags)
  *	QDDB_KEYLIST_FLAG_MARKFROMDATATREE:	Mark up a DataTree (args == DataTree *)
  *	QDDB_KEYLIST_FLAG_PRUNEBYROW:
  *	QDDB_KEYLIST_FLAG_PRUNEBYATTR:
+ *	QDDB_KEYLIST_FLAG_PRUNEBYATTRINT:
  */
 void *Qddb_KeyListProcess(schema, List, args, op, flags)
     Schema		*schema;
@@ -130,15 +145,21 @@ void *Qddb_KeyListProcess(schema, List, args, op, flags)
     List = KeyListGetList(&List, flags);
     switch (op) {
     case QDDB_KEYLIST_PROC_MARK:
-	if ((flags & QDDB_KEYLIST_FLAG_MARKFROMDATATREE) != 0)
+	if ((flags & QDDB_KEYLIST_FLAG_MARKFROMDATATREE) != 0) {
 	    return NULL;
-	else if ((flags & QDDB_KEYLIST_FLAG_MARKFROMINCORE) != 0)
+	} else if ((flags & QDDB_KEYLIST_FLAG_MARKFROMINCORE) != 0) {
 	    retval = KeyListMarkEntry(schema, List, (InCoreEntry *)args);
+	} else if ((flags & QDDB_KEYLIST_FLAG_MARKASANY) != 0) {
+	    KeyListMarkAsAny(List);
+	    retval = List;
+	}
 	break;
     case QDDB_KEYLIST_PROC_PRUNE:
-	if ((flags & QDDB_KEYLIST_FLAG_PRUNEBYATTR) != 0)
+	if ((flags & QDDB_KEYLIST_FLAG_PRUNEBYATTR) != 0) {
 	    retval = PruneListByAttribute(schema, List, (char *)args);
-	else if ((flags & QDDB_KEYLIST_FLAG_PRUNEBYROW) != 0) {
+	} else if ((flags & QDDB_KEYLIST_FLAG_PRUNEBYATTRINT) != 0) {
+	    retval = PruneListByIntegerAttribute(schema, List, *(int *)args);
+	} else if ((flags & QDDB_KEYLIST_FLAG_PRUNEBYROW) != 0) {
 	    prune_args = (Qddb_PruneArgs *)args;
 	    retval = KeyListPruneByRow(schema, List, prune_args->attrs, prune_args->attr_len);
 	} else {
@@ -259,7 +280,7 @@ static KeyList *KeyListGetList(node, flags)
     retval = NULL;
     while (*ptr != NULL) {
 	tmp = KeyListGetNode(&ptr, flags);
-	KeyListAddToList(&hashtable[QDDB_KEYLIST_HASH_VAL(tmp->Start)], tmp, flags);
+	KeyListAddToList(&hashtable[QDDB_KEYLIST_HASH_VAL(tmp->Start + tmp->Number)], tmp, flags);
     }
     retval = KeyListHashTable(hashtable);
     return retval;
@@ -275,7 +296,7 @@ static KeyList **KeyListBuildHashTable(List, flags)
     ptr = List;
     while (*ptr != NULL) {
 	tmp = KeyListGetNode(&ptr, flags);
-	KeyListAddToList(&retval[QDDB_KEYLIST_HASH_VAL(tmp->Start)], tmp, flags);
+	KeyListAddToList(&retval[QDDB_KEYLIST_HASH_VAL(tmp->Start + tmp->Number)], tmp, flags);
     }
     return retval;
 }
@@ -591,6 +612,23 @@ static KeyList *KeyListUnion(schema, List1, List2, flags)
     return retval;
 }
 
+static KeyList *KeyListConcat(schema, List1, List2, flags)
+    Schema		*schema;
+    KeyList		*List1, *List2;
+    int			flags;
+{
+    KeyList		*retval;
+
+    if (List1 == NULL)
+	return (KeyList *)Qddb_KeyListProcess(schema, List2, NULL, QDDB_KEYLIST_PROC_NULLOP, flags);
+    if (List2 == NULL)
+	return (KeyList *)Qddb_KeyListProcess(schema, List1, NULL, QDDB_KEYLIST_PROC_NULLOP, flags);
+    List1 = (KeyList *)Qddb_KeyListProcess(schema, List1, NULL, QDDB_KEYLIST_PROC_NULLOP, flags);
+    List2 = (KeyList *)Qddb_KeyListProcess(schema, List2, NULL, QDDB_KEYLIST_PROC_NULLOP, flags);
+    retval = Qddb_KeyListConcat(List1, List2, NULL);
+    return retval;
+}
+
 static KeyList *KeyListIntersection(schema, List1, List2, flags)
     Schema		*schema;
     KeyList		*List1, *List2;
@@ -650,9 +688,10 @@ static int CompareKeyNodesByRow(schema, Node1, Node2)
     Schema		*schema;
     KeyList		*Node1, *Node2;
 {
-    int			i, j, numdots, attr1, attr2;
+    int			i;
+    size_t		*inst1, *inst2, attr1, attr2;
+    int			min;
     SchemaNode		*node1, *node2;
-    char		buf1[BUFSIZ], buf2[BUFSIZ], *tmpbuf;
 
     if (Node1 == NULL)
 	return 1;
@@ -668,45 +707,34 @@ static int CompareKeyNodesByRow(schema, Node1, Node2)
     /* Find the number of places in the lowest common
      * ancestor.
      */
-    node1 = schema->Entries+attr1;
-    node2 = schema->Entries+attr2;
-    if (node1->AncestorNumber[1] != node2->AncestorNumber[1]) {
-	return 0; /* highest level mismatches always match in a join */
-    } 
-    for (i = 2; i < MIN(node1->Level, node2->Level); i++) {
+    node1 = (schema->Entries)+attr1;
+    node2 = (schema->Entries)+attr2;
+    inst1 = (size_t *)alloca(MAXIMUM_NUMBER_OF_SCHEMA_LEVELS*sizeof(size_t));
+    inst2 = (size_t *)alloca(MAXIMUM_NUMBER_OF_SCHEMA_LEVELS*sizeof(size_t));
+    StringToInstance(Node1->Instance, inst1);
+    StringToInstance(Node2->Instance, inst2);
+    min = MIN(node1->Level, node2->Level);
+    for (i = 1; i < min; i++) {
 	if (node1->AncestorNumber[i] != node2->AncestorNumber[i]) {
-	    break;
+	    return 0; /* attribute mismatches always match in a join */
+	}
+	if (inst1[i-1] != inst2[i-1]) {
+	    return inst1[i-1] - inst2[i-1];
 	}
     }
-    numdots = i-1;
-    /* Instance numbers must match for the number of places in
-     * the lowest common ancestor.
-     */
-    strcpy(buf1, Node1->Instance);
-    strcpy(buf2, Node2->Instance);
-    tmpbuf = buf1;
-    j = 0;
-    while (*tmpbuf != '\0') {
-	if (*tmpbuf == '.') {
-	    j++;
-	    if (j == numdots) {
-		*tmpbuf = '\0';
-	    }
+    if (node1->Level == min) {
+	if (node2->Level == min) {
+	    if (node1->Number != node2->Number)
+		return 0;
+	} else {
+	    if (node1->Number != node2->AncestorNumber[min])
+		return 0;
 	}
-	tmpbuf++;
+    } else {
+	if (node1->AncestorNumber[min] != node2->Number)
+	    return 0;
     }
-    tmpbuf = buf2;
-    j = 0;
-    while (*tmpbuf != '\0') {
-	if (*tmpbuf == '.') {
-	    j++;
-	    if (j == numdots) {
-		*tmpbuf = '\0';
-	    }
-	}
-	tmpbuf++;
-    }
-    return strcmp(buf1, buf2);
+    return inst1[min] - inst2[min];
 }
 
 static int MarkPaths(schema, Node, KeyMatrix, MarkedMatrix, AttrArray, KeyTop, AttrLen)
@@ -717,13 +745,15 @@ static int MarkPaths(schema, Node, KeyMatrix, MarkedMatrix, AttrArray, KeyTop, A
     int			i, FoundOne;
     KeyList		*tmp;
 
-    if (AttrLen == 0)
+    if (AttrLen == 0) {
 	return True;
+    }
     FoundOne = False;
     for (i = 0; i < KeyTop[0]; i++) {
 	tmp = Node;
-	while (tmp != NULL && CompareKeyNodesByRow(schema, tmp, KeyMatrix[0][i]) == 0)
+	while (tmp != NULL && CompareKeyNodesByRow(schema, tmp, KeyMatrix[0][i]) == 0) {
 	    tmp = tmp->next;
+	}
 	if (tmp == NULL) {
 	    KeyMatrix[0][i]->next = Node;
 	    if (MarkPaths(schema, KeyMatrix[0][i], KeyMatrix+1, MarkedMatrix+1, 
@@ -744,12 +774,16 @@ static KeyList *ProcessListByAttr(schema, List, Length, AttrList, AttrLen)
     char		*AttrList[];
     int			AttrLen;
 {
-    int			*AttrArray = (int *)Malloc(sizeof(int)*AttrLen);
-    int			*KeyTop    = (int *)Calloc(sizeof(int)*AttrLen);
-    KeyList		***KeyMatrix = (KeyList ***)Malloc(sizeof(KeyList **)*AttrLen), *TmpList;
-    int			**MarkedMatrix = (int **)Malloc(sizeof(int *)*AttrLen);
+    int			*AttrArray;
+    int			*KeyTop;
+    KeyList		***KeyMatrix, *TmpList;
+    int			**MarkedMatrix;
     int			i, j;
 
+    AttrArray    = (int *)Malloc(sizeof(int)*AttrLen);
+    KeyTop       = (int *)Calloc(sizeof(int)*AttrLen);
+    KeyMatrix    = (KeyList ***)Malloc(sizeof(KeyList **)*AttrLen);
+    MarkedMatrix = (int **)Malloc(sizeof(int *)*AttrLen);
     for (i = 0; i < AttrLen; i++)
 	KeyMatrix[i] = (KeyList **)Malloc(sizeof(KeyList *)*Length);
     for (i = 0; i < AttrLen; i++)
@@ -793,7 +827,8 @@ static KeyList *ProcessListByAttr(schema, List, Length, AttrList, AttrLen)
 		    TmpList = TmpList->next;
 		}
 	    } else {
-		Free(KeyMatrix[i][j]);
+		KeyMatrix[i][j]->next = NULL;
+		Qddb_Free(QDDB_TYPE_KEYLIST, KeyMatrix[i][j]);
 	    }
 	}
 	Free(KeyMatrix[i]);
@@ -874,9 +909,6 @@ static KeyList *KeyListPruneByRow(schema, List, AttrList, AttrLen)
     return List;
 }
 
-
-
-
 static void KeyListMarkAttrs(schema, List, ThisEntry)
     Schema		*schema;
     KeyList		*List;
@@ -970,7 +1002,6 @@ static KeyList *KeyListMarkEntry(schema, List, ThisEntry)
     return retval;
 }
 
-#if defined(notyet)
 static void KeyListMarkAsAny(list)
     KeyList		*list;
 {
@@ -979,7 +1010,6 @@ static void KeyListMarkAsAny(list)
 	list = list->next;
     }
 }
-#endif
 
 static KeyList *PruneListByAttribute(schema, list, attr)
     Schema			*schema;
@@ -1017,3 +1047,34 @@ static KeyList *PruneListByAttribute(schema, list, attr)
     }
     return retval;
 }
+
+static KeyList *PruneListByIntegerAttribute(schema, list, attr_num)
+    Schema			*schema;
+    KeyList			*list;
+    int				attr_num;
+{
+    KeyList			*last_place, *place, *retval;
+
+    last_place = NULL;
+    place = retval = list;
+    while (place != NULL) {
+	if (place->Attribute != attr_num) {
+	    /* prune this KeyList node */
+	    if (last_place == NULL) {
+		retval = place->next;
+		Free(place->Instance);
+		Free(place);
+		place = retval;
+	    } else {
+		last_place->next =  place->next;
+		Free(place->Instance);
+		Free(place);
+		place = last_place->next;
+	    }
+	} else {
+	    last_place = place;
+	    place = place->next;
+	}
+    }
+    return retval;
+}
