@@ -450,11 +450,13 @@ int Qddb_RowListQuery(schema, list, SepString,
     InCoreEntry		*CoreEntry = NULL;
     char		*RelationFN = schema->RelationName;
     int			DBFile, tmpattr;
-    size_t		*AttrNums = (size_t *)Malloc(sizeof(size_t)*AttrTop), i;
-    size_t		*AttrNumsUsed = (size_t *)Malloc(sizeof(size_t)*AttrsUsedTop);
+    size_t		*AttrNums, i;
+    size_t		*AttrNumsUsed;
     RowList		*Row;
     KeyList		**SplitList, **OrigSplitList, *tmplist, Key;
 
+    AttrNums = (size_t *)Malloc(sizeof(size_t)*AttrTop);
+    AttrNumsUsed = (size_t *)Malloc(sizeof(size_t)*AttrsUsedTop);
     if (list == NULL)
 	return 0;
     DBFile = schema->database_fd;
@@ -477,7 +479,7 @@ int Qddb_RowListQuery(schema, list, SepString,
 	AttrNumsUsed[i] = (size_t)tmpattr;
     }
     /* Here we copy the keylist into a split list; each element of the
-     * split list is deleted from MARKFROMINCORE
+     * split list is deleted by MARKFROMINCORE
      */
     (void)Qddb_KeyListProcess(schema, list, &SplitList, QDDB_KEYLIST_PROC_SPLIT, QDDB_KEYLIST_FLAG_COPY);
     OrigSplitList = SplitList;
@@ -599,7 +601,7 @@ static void QueryPrintRow(schema, Key, ThisEntry, Row, SepString,
 		    Qddb_ConcatBuffer(SepString);
 		} else {
 		    buf_ptr = Malloc(AttrWidth[i]+strlen(SepString)+2);
-		    sprintf(buf_ptr, "%*s%s", AttrWidth[i], " ", SepString);
+		    sprintf(buf_ptr, "%*s%s", (int)AttrWidth[i], " ", SepString);
 		    Qddb_ConcatBuffer(buf_ptr);
 		    Free(buf_ptr);
 		}
@@ -621,11 +623,11 @@ static void QueryPrintRow(schema, Key, ThisEntry, Row, SepString,
 	}
 	buf_ptr = Qddb_GetBuffer();
 	if (TCLprint == True) {
+#if defined(USE_TCL)
 	    char		key_buf[BUFSIZ], *row_buf = NULL, *merged_buf;
 	    int			nargc = 0;
 	    char		*nargv[4];
 
-#if defined(USE_TCL)
 	    nargv[nargc++] = buf_ptr;
 	    sprintf(key_buf, "%d %d %d %d",
 		    (int)Key->Start, (int)Key->Length, (int)Key->Number, (int)QDDB_KEYLIST_TYPE(Key));
@@ -633,6 +635,7 @@ static void QueryPrintRow(schema, Key, ThisEntry, Row, SepString,
 	    if (DontPrintRows == False) {
 		row_buf = nargv[nargc++] = PrintRowInstances(schema, ThisEntry, Row);
 	    }
+	    nargv[nargc] = NULL;
 	    merged_buf = Tcl_Merge(nargc, nargv);
 	    Free(buf_ptr);
 	    if (row_buf != NULL)
@@ -685,7 +688,7 @@ static void QueryPrintRowAsElements(interp, schema, Key, ThisEntry, Row, SepStri
 	}
 	thisrow = Row;
 	if (!suppress_printing) {
-	    merge_buf = (char **)Malloc(sizeof(char *)*AttrTop);
+	    merge_buf = (char **)Malloc(sizeof(char *)*(AttrTop+1));
 	    for (i = 0; i < AttrTop; i++) {
 		tmp = Row;
 		while (tmp != NULL) {
@@ -700,7 +703,7 @@ static void QueryPrintRowAsElements(interp, schema, Key, ThisEntry, Row, SepStri
 			*(merge_buf[i]) = '\0';
 		    } else {
 			buf_ptr = Malloc((size_t)AttrWidth[i]+1);
-			sprintf(buf_ptr, "%*s%s", AttrWidth[i], " ", SepString);
+			sprintf(buf_ptr, "%*s%s", (int)AttrWidth[i], " ", SepString);
 			merge_buf[i] = buf_ptr;
 		    }
 		} else {
@@ -716,6 +719,7 @@ static void QueryPrintRowAsElements(interp, schema, Key, ThisEntry, Row, SepStri
 		    }
 		}
 	    }
+	    merge_buf[AttrTop] = NULL;
 	    buf_ptr = Tcl_Merge((int)AttrTop, merge_buf);
 	    for (i = 0; i < AttrTop; i++)
 		Free(merge_buf[i]);
@@ -730,6 +734,7 @@ static void QueryPrintRowAsElements(interp, schema, Key, ThisEntry, Row, SepStri
 	if (DontPrintRows == False) {
 	    row_buf = nargv[nargc++] = PrintRowInstances(schema, ThisEntry, Row);
 	}
+	nargv[nargc] = NULL;
 	merged_buf = Tcl_Merge(nargc, nargv);
 	Free(buf_ptr);
 	if (row_buf != NULL)
