@@ -38,8 +38,9 @@ static KeyList *RegexpSearch _ANSI_ARGS_((Schema *, char *, size_t *, int));
 static KeyList *NumericSearch _ANSI_ARGS_((Schema *, double, size_t *, int));
 static KeyList *NumericRangeSearch _ANSI_ARGS_((Schema *, NumericRange *, size_t *, int));
 static KeyList *StringRangeSearch _ANSI_ARGS_((Schema *, StringRange *, size_t *, int));
-static KeyList *Search _ANSI_ARGS_((Schema *, char *, size_t *, int));
+static KeyList *Search _ANSI_ARGS_((Schema *, char *, size_t *, int, int));
 static KeyList *SecondarySearch _ANSI_ARGS_((Schema *, size_t *, KeyList *, Qddb_SearchArg *, int));
+static KeyList *CachedSecondarySearch _ANSI_ARGS_((Schema *, size_t *, KeyList *, Qddb_SearchArg *, int));
 
 static KeyList *IndexLinesToKeyList _ANSI_ARGS_((Schema *, Entry, size_t *, int));
 
@@ -67,7 +68,11 @@ KeyList *Qddb_Search(schema, arg, number_read, attrnum)
 	retval = NumericSearch(schema, arg->Number, number_read, attrnum);
 	break;
     case ALPHA:
-	retval = Search(schema, arg->Alpha, number_read, attrnum);
+    case ALPHAPARSE:
+	retval = Search(schema, arg->Alpha, number_read, attrnum, 1);
+	break;
+    case ALPHANOPARSE:
+	retval = Search(schema, arg->Alpha, number_read, attrnum, 0);
 	break;
     case NUMERICRANGE:
 	NumericRangeVar.UpperExists = arg->UpperExists;
@@ -87,6 +92,10 @@ KeyList *Qddb_Search(schema, arg, number_read, attrnum)
 	qddb_errno = QDDB_ERRNO_INVALID_ARG;
 	return NULL;
     }
+#if defined(notyet)
+    if (attrnum == -1)
+	(void)Qddb_KeyListProcess(schema, retval, NULL, QDDB_KEYLIST_PROC_MARK, QDDB_KEYLIST_FLAG_MARKASANY);
+#endif
     return retval;
 }
 
@@ -131,13 +140,13 @@ static KeyList *RegexpSearch(schema, Regexp, NumberRead, attrnum)
 	return NULL;
     if (KeyFile == -1 || IndexFile == -1 || SizeOfFile(KeyFile) == 0 || SizeOfFile(IndexFile) == 0)
 	goto secondary_search;
-    Chunk = Malloc(CHUNKSIZE+1);
+    Chunk = Malloc(CHUNK_SIZE+1);
     /* Read in a chunk at a time, parse it up into newline separated
      * pieces and compare the regular expression to each word.
      */
     Size = SizeOfFile(KeyFile);
-    NumberOfChunks = (Size/CHUNKSIZE);
-    if (Size % CHUNKSIZE != 0)
+    NumberOfChunks = (Size/CHUNK_SIZE);
+    if (Size % CHUNK_SIZE != 0)
 	NumberOfChunks++;
 #ifdef DEBUGREGEXP
     fprintf(stderr, "RegexpSearch: %d chunks\n", NumberOfChunks);
@@ -148,7 +157,7 @@ static KeyList *RegexpSearch(schema, Regexp, NumberRead, attrnum)
 	fprintf(stderr, "RegexpSearch: %dth chunk (top of loop)\n", i);
 	fflush(stderr);
 #endif
-	if ((NumberOfBytes = read(KeyFile, Chunk, CHUNKSIZE)) == -1)
+	if ((NumberOfBytes = read(KeyFile, Chunk, CHUNK_SIZE)) == -1)
 	    PANIC("RegexpSearch on read");
 	else {
 	    char		*lastnewline;
@@ -207,14 +216,20 @@ static KeyList *RegexpSearch(schema, Regexp, NumberRead, attrnum)
 		    while (*strend != ' ')	/* hit a space */
 			strend++;
 		    *strend = '\0';
-		    offset = atoi(stroffset);
+		    if (schema->UseCondensedIndexing == False)
+			offset = Qddb_BaseToOffset(stroffset, 10);
+		    else
+			offset = Qddb_BaseToOffset(stroffset, 62);
 		    *strend++ = ' ';
 		    /* length */
 		    strlength = strend;
 		    while (*strend != '\n')	/* hit the newline */
 			strend++;
 		    *strend = '\0';
-		    length = atoi(strlength);
+		    if (schema->UseCondensedIndexing == False)
+			length = Qddb_BaseToSize(strlength, 10);
+		    else
+			length = Qddb_BaseToSize(strlength, 62);
 		    *strend++ = '\n';
 		    /* read in the Index entry */
 		    lseek(IndexFile, offset, SEEK_SET);
@@ -264,7 +279,8 @@ secondary_search:
 
 static void Chunking_GetChunk _ANSI_ARGS_((int, char *, off_t, size_t));
 static void Chunking_AdjustBounds _ANSI_ARGS_((char **, ChunkBounds *, size_t, size_t));
-static ChunkBounds *Chunking_Search _ANSI_ARGS_((int, Qddb_SearchArg *, ChunkBounds *, size_t));
+static ChunkBounds *Chunking_Search _ANSI_ARGS_((int, Qddb_SearchArg *, \
+						 ChunkBounds *, ChunkBounds *, int, int));
 static void Chunking_FixBounds _ANSI_ARGS_((int, ChunkBounds *, size_t, size_t, Qddb_SearchArg *));
 static char *Chunking_Ending _ANSI_ARGS_((char *, size_t));
 static char *Chunking_Beginning _ANSI_ARGS_((char *, size_t));
@@ -309,14 +325,6 @@ static char *Chunking_Beginning(buf, len)
     *ch = '\0';
     retval = Qddb_StripKey(buf+1, newlen);
     *ch = ' ';
-#if defined(notdef)
-    for (i = 1; i < len; i++) {
-	if (buf[i] == ' ')
-	    break;
-    }
-    retval = Malloc((size_t)i);
-    strncpy(retval, buf+1, (size_t)i-2);
-#endif
     return retval;
 }
 
@@ -332,7 +340,7 @@ static char *Chunking_Ending(buf, len)
     if (buf[last] != '\n')
 	fprintf(stderr, "Chunking_Ending: buf[%d] != '\\n'", (int)last);
 #endif
-    for (i = last-1; i > 0; i--) {
+    for (i = last; i >= 0; i--) {
 	if (buf[i] == '\n')
 	    break;
     }
@@ -342,14 +350,6 @@ static char *Chunking_Ending(buf, len)
     *ch = '\0';
     retval = Qddb_StripKey(buf+i, newlen);
     *ch = ' ';
-#if defined(notdef)
-    retval = Malloc((size_t)(last-i));
-    for (; i < last; i++) {
-	if (buf[i] == ' ')
-	    break;
-    }
-    strncpy(retval, buf+first, i-first);
-#endif
     return retval;
 }
 
@@ -383,14 +383,14 @@ static void Chunking_AdjustBounds(obuf, bounds, mid, num)
 	while (*buf != '\n') {
 	    bounds[mid].start++;
 	    bounds[mid].length--;
-	    if (mid != 0) {
+	    if (mid > 0) {
 		bounds[mid-1].length++;
 	    }
 	    buf++;
 	}
 	while (*endbuf != '\n') {
 	    bounds[mid].length--;
-	    if (mid != num-1) {
+	    if (mid < num-1) {
 		bounds[mid+1].start--;
 		bounds[mid+1].length++;
 	    }
@@ -400,26 +400,27 @@ static void Chunking_AdjustBounds(obuf, bounds, mid, num)
     *obuf = buf;
 }
 
-static ChunkBounds *Chunking_Search(fd, search_arg, bounds, num)
+static ChunkBounds *Chunking_Search(fd, search_arg, obounds, bounds, onum, num)
     int			fd;
     Qddb_SearchArg	*search_arg;
-    ChunkBounds		*bounds;
-    size_t		num;
+    ChunkBounds		*obounds, *bounds;
+    int			onum, num;
 {
-    size_t		mid = num/2;
+    int			mid = num/2;
     off_t		start;
     size_t		length;
     char		*buf, *ending, *beginning;
     double		end_num, begin_num;
     ChunkBounds		*retval;
 
-    if (num <= 0)
+    if (num <= 0) {
 	return NULL;
+    }
     start = bounds[mid].start;
     length = bounds[mid].length;
-    buf = alloca(length);
+    buf = (char *)alloca(length*2);
     Chunking_GetChunk(fd, buf, start, length);
-    Chunking_AdjustBounds(&buf, bounds, mid, num);
+    Chunking_AdjustBounds(&buf, obounds, (size_t)((bounds - obounds) + mid), (size_t)onum);
     length = bounds[mid].length;
     /* Check the extremities to see if the desired number matches either
      * endpoint or is contained in the middle somewhere.
@@ -427,14 +428,14 @@ static ChunkBounds *Chunking_Search(fd, search_arg, bounds, num)
     beginning = Chunking_Beginning(buf, length);
     ending = Chunking_Ending(buf, length);
     if (search_arg->Type == NUMERIC) {
-	begin_num = strtod(beginning, NULL);
-	end_num = strtod(ending, NULL);
+	begin_num = Qddb_GetNum(beginning);
+	end_num = Qddb_GetNum(ending);
 	if (search_arg->Number < begin_num) {
-	    retval = Chunking_Search(fd, search_arg, bounds, mid);
+	    retval = Chunking_Search(fd, search_arg, obounds, bounds, onum, mid);
 	    if (retval == NULL)
 		retval = bounds+mid;
 	} else if (search_arg->Number > end_num) {
-	    retval = Chunking_Search(fd, search_arg, bounds+mid+1, num-mid-1);
+	    retval = Chunking_Search(fd, search_arg, obounds, bounds+mid+1, onum, num-mid-1);
 	    if (retval == NULL)
 		retval = bounds+mid;
 	} else {
@@ -445,11 +446,11 @@ static ChunkBounds *Chunking_Search(fd, search_arg, bounds, num)
 	}
     } else {
 	if (strcmp(search_arg->Alpha, beginning) < 0) {
-	    retval = Chunking_Search(fd, search_arg, bounds, mid);
+	    retval = Chunking_Search(fd, search_arg, obounds, bounds, onum, mid);
 	    if (retval == NULL)
 		retval = bounds+mid;
 	} else if (strcmp(search_arg->Alpha, ending) > 0) {
-	    retval = Chunking_Search(fd, search_arg, bounds+mid+1, num-mid-1);
+	    retval = Chunking_Search(fd, search_arg, obounds, bounds+mid+1, onum, num-mid-1);
 	    if (retval == NULL)
 		retval = bounds+mid;
 	} else {
@@ -479,7 +480,7 @@ static void Chunking_FixBounds(fd, bounds, idx, num, arg)
 	Chunking_AdjustBounds(&buf, bounds,  (size_t)i, num);
 	ending = Chunking_Ending(buf, bounds[i].length);
 	if (arg->Type == NUMERIC) {
-	    if (arg->Number > strtod(ending, NULL)) {
+	    if (arg->Number > Qddb_GetNum(ending)) {
 		Free(ending);
 		Free(obuf);
 		break;
@@ -502,7 +503,7 @@ static void Chunking_FixBounds(fd, bounds, idx, num, arg)
 	Chunking_AdjustBounds(&buf, bounds, (size_t)i, num);
 	beginning = Chunking_Beginning(buf, bounds[i].length);
 	if (arg->Type == NUMERIC) {
-	    if (arg->Number < strtod(beginning, NULL)) {
+	    if (arg->Number < Qddb_GetNum(beginning)) {
 		Free(beginning);
 		Free(obuf);
 		break;
@@ -522,7 +523,7 @@ static void Chunking_FixBounds(fd, bounds, idx, num, arg)
 }
 
 /* IndexLinesToKeyList -- Convert an Entry full of *Index lines
- * to a single KeyList.   This routine destroys the Entry 'lines.'
+ * to a single KeyList.
  */
 static KeyList *IndexLinesToKeyList(schema, lines, num_read, attrnum)
     Schema		*schema;
@@ -566,8 +567,13 @@ static KeyList *IndexLinesToKeyList(schema, lines, num_read, attrnum)
 	    ch++;
 	*ch++ = '\0';
 	length_ch = ch;
-	start = atoi(start_ch);
-	length = atoi(length_ch);
+	if (schema->UseCondensedIndexing == False) {
+	    start = Qddb_BaseToOffset(start_ch, 10);
+	    length = Qddb_BaseToSize(length_ch, 10);
+	} else {
+	    start = Qddb_BaseToOffset(start_ch, 62);
+	    length = Qddb_BaseToSize(length_ch, 62);
+	}
 	buf = Malloc((size_t)length+1);
 	buf[length] = '\0';
 	if (lseek(fd, (off_t)start, SEEK_SET) < 0) 
@@ -621,7 +627,7 @@ static Entry PruneLines(lines, lower, upper)
 	/* Throw-aways */
 	if (lower != NULL) {
 	    while (*tmp != NULL) {
-		if (lower->Number <= strtod(*tmp, NULL))
+		if (lower->Number <= Qddb_GetNum(*tmp))
 		    break;
 		Free(*tmp);
 		tmp++;
@@ -631,7 +637,7 @@ static Entry PruneLines(lines, lower, upper)
 	top = tmp;
 	while (*tmp != NULL) {
 	    i++;
-	    if (upper != NULL && upper->Number < strtod(*tmp, NULL))
+	    if (upper != NULL && upper->Number < Qddb_GetNum(*tmp))
 		break;
 	    tmp++;
 	}
@@ -705,7 +711,7 @@ static KeyList *NumericSearch(schema, number, number_read, attrnum)
     int			attrnum;
 {
     int			numidx_fd;
-    char		*numidx_path = alloca(strlen(schema->RelationName)+14), *buf, *obuf;
+    char		*numidx_path = (char *)alloca(strlen(schema->RelationName)+14), *buf, *obuf;
     ChunkBounds		*chunk_bounds, *find_bounds;
     int			idx, minidx, maxidx, i;
     off_t		buf_start;
@@ -721,7 +727,8 @@ static KeyList *NumericSearch(schema, number, number_read, attrnum)
     if ((numidx_fd = open(numidx_path, O_RDONLY)) < 0 || SizeOfFile(numidx_fd) == 0)
 	goto secondary_search;
     Chunking_Build(numidx_fd, CHUNK_SIZE, &chunk_bounds, &numchunks);
-    find_bounds = Chunking_Search(numidx_fd, &search_arg, chunk_bounds, numchunks);
+    find_bounds = Chunking_Search(numidx_fd, &search_arg, chunk_bounds, 
+				  chunk_bounds, (int)numchunks, (int)numchunks);
     idx = find_bounds - chunk_bounds;
     chunk_bounds[idx].used = True;
     Chunking_FixBounds(numidx_fd, chunk_bounds, (size_t)idx, numchunks, &search_arg);
@@ -736,10 +743,11 @@ static KeyList *NumericSearch(schema, number, number_read, attrnum)
 	buf_len += chunk_bounds[i].length;
     }
     Free(chunk_bounds);
-    obuf = buf = Malloc(buf_len);
+    obuf = buf = Malloc(buf_len+1);
+    buf[buf_len] = '\n';
     Chunking_GetChunk(numidx_fd, buf, buf_start, buf_len);
     lines = NULL;
-    Qddb_SplitBufferIntoLines(&lines, buf, buf_len);
+    Qddb_SplitBufferIntoLines(&lines, buf, buf_len+1);
     lines = PruneLines(lines, &search_arg, &search_arg);
     retval = IndexLinesToKeyList(schema, lines, number_read, attrnum);
     Qddb_Free(QDDB_TYPE_ENTRY, lines);
@@ -766,7 +774,7 @@ static KeyList *NumericRangeSearch(schema, nr_var, number_read, attrnum)
     int			attrnum;
 {
     int			numidx_fd;
-    char		*numidx_path = alloca(strlen(schema->RelationName)+14), *buf, *obuf;
+    char		*numidx_path = (char *)alloca(strlen(schema->RelationName)+14), *buf, *obuf;
     ChunkBounds		*chunk_bounds, *find_bounds;
     int			idx = 0, idx2 = 0, minidx, maxidx, i;
     off_t		buf_start;
@@ -783,7 +791,8 @@ static KeyList *NumericRangeSearch(schema, nr_var, number_read, attrnum)
     if (nr_var->LowerExists == True) {
 	search_arg.Type = NUMERIC;
 	search_arg.Number = nr_var->Lower;
-	find_bounds = Chunking_Search(numidx_fd, &search_arg, chunk_bounds, numchunks);
+	find_bounds = Chunking_Search(numidx_fd, &search_arg, chunk_bounds, chunk_bounds, 
+				      (int)numchunks, (int)numchunks);
 	idx = find_bounds - chunk_bounds;
 	chunk_bounds[idx].used = True;
 	Chunking_FixBounds(numidx_fd, chunk_bounds, (size_t)idx, numchunks, &search_arg);
@@ -791,7 +800,8 @@ static KeyList *NumericRangeSearch(schema, nr_var, number_read, attrnum)
     if (nr_var->UpperExists == True) {
 	search_arg.Type = NUMERIC;
 	search_arg.Number = nr_var->Upper;
-	find_bounds = Chunking_Search(numidx_fd, &search_arg, chunk_bounds, numchunks);
+	find_bounds = Chunking_Search(numidx_fd, &search_arg, chunk_bounds, chunk_bounds, 
+				      (int)numchunks, (int)numchunks);
 	idx2 = find_bounds - chunk_bounds;
 	chunk_bounds[idx2].used = True;
 	Chunking_FixBounds(numidx_fd, chunk_bounds, (size_t)idx2, numchunks, &search_arg);
@@ -826,10 +836,11 @@ static KeyList *NumericRangeSearch(schema, nr_var, number_read, attrnum)
 	buf_len += chunk_bounds[i].length;
     }
     Free(chunk_bounds);
-    obuf = buf = Malloc(buf_len);
+    obuf = buf = Malloc(buf_len+1);
+    buf[buf_len] = '\n';
     Chunking_GetChunk(numidx_fd, buf, buf_start, buf_len);
     lines = NULL;
-    Qddb_SplitBufferIntoLines(&lines, buf, buf_len);
+    Qddb_SplitBufferIntoLines(&lines, buf, buf_len+1);
     lower_arg.Type = NUMERIC;
     lower_arg.Number = nr_var->Lower;
     upper_arg.Type = NUMERIC;
@@ -837,6 +848,7 @@ static KeyList *NumericRangeSearch(schema, nr_var, number_read, attrnum)
     lines = PruneLines(lines, nr_var->LowerExists == True? &lower_arg:NULL, 
 		       nr_var->UpperExists == True? &upper_arg:NULL);
     retval = IndexLinesToKeyList(schema, lines, number_read, attrnum);
+    Free(obuf);
     Qddb_Free(QDDB_TYPE_ENTRY, lines);
     /* buf contains each row with 'number' in it. Now we need to parse up 
      * the rows, read the 'Index' entries, and build the KeyLists.
@@ -864,7 +876,7 @@ static KeyList *StringRangeSearch(schema, sr_var, number_read, attrnum)
     int			attrnum;
 {
     int			stridx_fd;
-    char		*stridx_path = alloca(strlen(schema->RelationName)+10), *buf, *obuf, *ch;
+    char		*stridx_path = (char *)alloca(strlen(schema->RelationName)+10), *buf, *obuf, *ch;
     ChunkBounds		*chunk_bounds, *find_bounds;
     int			idx = 0, idx2 = 0, minidx, maxidx, i;
     off_t		buf_start;
@@ -872,7 +884,7 @@ static KeyList *StringRangeSearch(schema, sr_var, number_read, attrnum)
     Entry		lines;
     KeyList		*retval = NULL;
     Qddb_SearchArg	search_arg, lower_arg, upper_arg;
-    
+
     strcpy(stridx_path, schema->RelationName);
     strcat(stridx_path, "/KeyIndex");
     if ((stridx_fd = open(stridx_path, O_RDONLY)) < 0 || SizeOfFile(stridx_fd) == 0)
@@ -883,7 +895,8 @@ static KeyList *StringRangeSearch(schema, sr_var, number_read, attrnum)
 	    *ch = tolower(*ch);
 	search_arg.Type = ALPHA;
 	search_arg.Alpha = sr_var->Lower;
-	find_bounds = Chunking_Search(stridx_fd, &search_arg, chunk_bounds, numchunks);
+	find_bounds = Chunking_Search(stridx_fd, &search_arg, chunk_bounds, chunk_bounds, 
+				      (int)numchunks, (int)numchunks);
 	idx = find_bounds - chunk_bounds;
 	chunk_bounds[idx].used = True;
 	Chunking_FixBounds(stridx_fd, chunk_bounds, (size_t)idx, numchunks, &search_arg);
@@ -893,7 +906,8 @@ static KeyList *StringRangeSearch(schema, sr_var, number_read, attrnum)
 	    *ch = tolower(*ch);
 	search_arg.Type = ALPHA;
 	search_arg.Alpha = sr_var->Upper;
-	find_bounds = Chunking_Search(stridx_fd, &search_arg, chunk_bounds, numchunks);
+	find_bounds = Chunking_Search(stridx_fd, &search_arg, chunk_bounds, chunk_bounds, 
+				      (int)numchunks, (int)numchunks);
 	idx2 = find_bounds - chunk_bounds;
 	chunk_bounds[idx2].used = True;
 	Chunking_FixBounds(stridx_fd, chunk_bounds, (size_t)idx2, numchunks, &search_arg);
@@ -928,10 +942,11 @@ static KeyList *StringRangeSearch(schema, sr_var, number_read, attrnum)
 	buf_len += chunk_bounds[i].length;
     }
     Free(chunk_bounds);
-    obuf = buf = Malloc(buf_len);
+    obuf = buf = Malloc(buf_len+1);
+    buf[buf_len] = '\n'; /* Qddb_SplitBufferIntoLines expects this */
     Chunking_GetChunk(stridx_fd, buf, buf_start, buf_len);
     lines = NULL;
-    Qddb_SplitBufferIntoLines(&lines, buf, buf_len);
+    Qddb_SplitBufferIntoLines(&lines, buf, buf_len+1);
     lower_arg.Type = ALPHA;
     lower_arg.Alpha = sr_var->Lower;
     upper_arg.Type = ALPHA;
@@ -939,6 +954,7 @@ static KeyList *StringRangeSearch(schema, sr_var, number_read, attrnum)
     lines = PruneLines(lines, sr_var->LowerExists == True? &lower_arg:NULL, 
 		       sr_var->UpperExists == True? &upper_arg:NULL);
     retval = IndexLinesToKeyList(schema, lines, number_read, attrnum);
+    Free(obuf);
     Qddb_Free(QDDB_TYPE_ENTRY, lines);
 secondary_search:
     search_arg.Type = ALPHARANGE;
@@ -953,11 +969,11 @@ secondary_search:
    
 #endif /* SQL */
 
-static KeyList *Search(schema, String, NumberRead, attrnum)
+static KeyList *Search(schema, String, NumberRead, attrnum, parsing)
     Schema		*schema;
     char        	*String;
     size_t	    	*NumberRead;
-    int			attrnum;
+    int			attrnum, parsing;
 {
     char		*RelationFN = schema->RelationName;
     char        	KeyFN[MAXFILENAMELEN]; 
@@ -985,7 +1001,7 @@ fprintf(stderr, "Search()\n");
     teststr = (char *)Malloc(strlen(String)+1);
     strcpy(teststr, String);
     *NumberRead = 0;
-    if (attrnum == -1) {
+    if (attrnum == -1 && parsing) {
 	last = 0;
 	for (i = 0; i < len; i++) {
 	    if (ispunct(teststr[i]) || isspace(teststr[i])) {
@@ -994,18 +1010,18 @@ fprintf(stderr, "Search()\n");
 		    continue;
 		}
 		if (ReturnValue == NULL) {
-		    ReturnValue = Search(schema, teststr+last, &Length, attrnum);
+		    ReturnValue = Search(schema, teststr+last, &Length, attrnum, parsing);
 		    *NumberRead += Length;
 		} else {
 		    ReturnValue = Qddb_KeyListOp(schema, 
-						 Search(schema, teststr+last, &Length, attrnum), ReturnValue,
-						 QDDB_KEYLIST_OP_INTERSECTION, 0);
+						 Search(schema, teststr+last, &Length, attrnum, parsing),
+						 ReturnValue, QDDB_KEYLIST_OP_INTERSECTION, 0);
 		    *NumberRead += Length;
 		}
 		last = i+1;
 	    }
 	}
-    } else {
+    } else if (parsing) {
 	size_t			place_in_string;
 	size_t			teststr_len = strlen(teststr);
 
@@ -1015,11 +1031,11 @@ fprintf(stderr, "Search()\n");
 	    while (tok_str != NULL) {
 		place_in_string += strlen(tok_str) + 1;
 		if (ReturnValue == NULL) {
-		    ReturnValue = Search(schema, tok_str, &Length, attrnum);
+		    ReturnValue = Search(schema, tok_str, &Length, attrnum, parsing);
 		    *NumberRead += Length;
 		} else {
 		    ReturnValue = Qddb_KeyListOp(schema, 
-						 Search(schema, tok_str, &Length, attrnum), 
+						 Search(schema, tok_str, &Length, attrnum, parsing), 
 						 ReturnValue, QDDB_KEYLIST_OP_INTERSECTION, 0);
 		    *NumberRead += Length;
 		}
@@ -1101,6 +1117,9 @@ KeyList *Qddb_ParseKey(schema, String, NumberRead, NormalIndex, attrnum)
     KeyList        	*new, *list = NULL;
     size_t		len;
 
+#if defined(DEBUG)
+    printf("Qddb_ParseKey(attr=%d)\n", attrnum);
+#endif
     place = Qddb_FindKey(String, &len) + 1;
     i = place - String;
     /* Build the key list */
@@ -1145,7 +1164,7 @@ KeyList *Qddb_ParseKey(schema, String, NumberRead, NormalIndex, attrnum)
 		    Free(word[3]);
 		}
 		continue;
-	    }
+	    } 
             num = 0;
             if (list == NULL) {
                 list = (KeyList *)Calloc(sizeof(KeyList));
@@ -1155,8 +1174,13 @@ KeyList *Qddb_ParseKey(schema, String, NumberRead, NormalIndex, attrnum)
                 new->next = list;
                 list = new;
             }
-	    list->Start = atoi(word[0]);
-	    list->Length = atoi(word[1]);
+	    if (schema->UseCondensedIndexing == False) {
+		list->Start = Qddb_BaseToOffset(word[0], 10);
+		list->Length = Qddb_BaseToSize(word[1], 10);
+	    } else {
+		list->Start = Qddb_BaseToOffset(word[0], 62);
+		list->Length = Qddb_BaseToSize(word[1], 62);
+	    }
 	    list->Attribute = attribute_number;
 	    if (schema->UseReducedAttributeIdentifiers == True) {
 		if (qddb_stabilizing == 1) {
@@ -1179,6 +1203,56 @@ KeyList *Qddb_ParseKey(schema, String, NumberRead, NormalIndex, attrnum)
     return list;
 }
 
+KeyList *Qddb_ParseKey2(schema, String, wordend)
+    Schema		*schema;
+    char        	*String, **wordend;
+{
+    int        		i, j, num, NUM;
+    char        	*place, *word[4];
+    KeyList        	*new, *list = NULL;
+    size_t		len;
+
+#if defined(DEBUG)
+    printf("Qddb_ParseKey2()\n");
+#endif
+    place = Qddb_FindKey(String, &len);
+    *wordend = place;
+    place++;
+    i = place - String;
+    /* Build the key list */
+    NUM = 4;
+    for (j = i+1, num = 0, word[0] = place; String[j] != '\0'; j++) {
+        if (String[j] == ' ' || String[j] == '\n') {
+            num++;
+	    if (num < NUM) {
+		word[num] = String + j + 1;
+	    }
+	    String[j] = '\0';
+	}
+	if (num == NUM) {
+            num = 0;
+            if (list == NULL) {
+                list = (KeyList *)Calloc(sizeof(KeyList));
+                list->next = NULL;
+            } else {
+                new = (KeyList *)Calloc(sizeof(KeyList));
+                new->next = list;
+                list = new;
+            }
+	    list->Start = 0;
+	    list->Length = 0;
+	    list->Type = atoi(word[0]);
+	    list->Number = atoi(word[1]);
+	    list->Attribute = atoi(word[2]);
+	    list->Instance = (char *)Malloc(strlen(word[3])+1);
+	    strcpy(list->Instance, word[3]);
+            place = String + j + 1;
+	    word[0] = place;
+        }
+    }
+    return list;
+}
+
 
 static KeyList *SecondarySearch(schema, NumberRead, ExistingList, Arg, attrnum)
     Schema		*schema;
@@ -1187,24 +1261,25 @@ static KeyList *SecondarySearch(schema, NumberRead, ExistingList, Arg, attrnum)
     Qddb_SearchArg	*Arg;
     int			attrnum;
 {
+    KeyList        	*NewList = NULL, *LastNode = NULL;
+    Entry		entry = NULL, orig_entry = NULL;
     char        	*RelationFN = schema->RelationName;
     char        	ChangeDir[MAXFILENAMELEN], AddDir[MAXFILENAMELEN];
     char        	TmpDir[MAXFILENAMELEN], dirnam[MAXFILENAMELEN];
-    char        	*Buffer, *Line, *Token;
-    KeyList        	*NewList = NULL, *LastNode = NULL;
-    int			dirlen;
-    size_t		attr;
+    char        	*Line, *Token;
+    int			dirlen, filedesc;
+    size_t		attr, len;
     char		*inst;
-    size_t	    	AddCount = 0, bufsize = BUFSIZ;
-    FILE        	*FilePtr;
+    size_t	    	AddCount = 0;
     DIR        		*dp, *opendir();
     struct dirent    	*DirEntry;
     
 #if defined(DEBUG)
-    fprintf(stderr, "in SecondarySearch\n");
-    fflush(stderr);
+    printf("in SecondarySearch\n");
+    fflush(stdout);
 #endif
-    Buffer = Line = (char *)Malloc(bufsize);
+    if (schema->UseCachedSecondarySearch == True)
+	return CachedSecondarySearch(schema, NumberRead, ExistingList, Arg, attrnum);
     strcpy(ChangeDir, RelationFN);
     strcat(ChangeDir, "/Changes/");
     strcpy(AddDir, RelationFN);
@@ -1220,21 +1295,16 @@ static KeyList *SecondarySearch(schema, NumberRead, ExistingList, Arg, attrnum)
 	if (dirlen == 0 || dirnam[0] == '.')
 		continue;
         sprintf(TmpDir, "%s%s", ChangeDir, dirnam);
-        FilePtr = fopen(TmpDir, "r");
-        if (FilePtr == NULL)
+        filedesc = open(TmpDir, O_RDONLY);
+        if (filedesc < 0)
             PANIC("An existing file in Changes/ cannot be opened");
-	fgets(Line, bufsize, FilePtr); /* get %0 */
-        while (fgets(Line, bufsize, FilePtr) != NULL) {
+	entry = orig_entry;
+	Qddb_ReadEntry(filedesc, &entry, (off_t)0, (size_t)0, True);
+	orig_entry = entry++;
+        while (*entry != NULL) {
             char        *StringPtr;
 
-	    while (Line[strlen(Line)-2] == '\\' || Line[strlen(Line)-1] != '\n') {
-		if (strlen(Line) >= bufsize/2) {
-		    bufsize *= 2;
-		    Buffer = Line = (char *)Realloc(Buffer, bufsize);
-		}
-		if (fgets(Line+strlen(Line), bufsize-strlen(Line), FilePtr) == NULL)
-		    break;
-	    }
+	    Line = *entry++;
 	    StringPtr = Line;
 	    while (*StringPtr != '\0') {
 		if (_isupper(*StringPtr))
@@ -1242,7 +1312,9 @@ static KeyList *SecondarySearch(schema, NumberRead, ExistingList, Arg, attrnum)
 		StringPtr++;
 	    }
 	    Token = strtok(Line, " "); /* Skip id */
-	    Line = Line+strlen(Token)+1;
+	    len = strlen(Token);
+	    len++;
+	    Line = Line+len;
 	    if (Qddb_TranslateExternalID(schema, Token, &attr, &inst) != 0) {
 		fprintf(stderr, "Warning: attribute %s malformed, skipping..\n", Token);
 		continue;
@@ -1270,14 +1342,14 @@ static KeyList *SecondarySearch(schema, NumberRead, ExistingList, Arg, attrnum)
 		    Add = True;
 		    break;
 	        case NUMERIC:
-		    if (strtod(Token, NULL) != Arg->Number)
+		    if (Qddb_GetNum(Token) != Arg->Number)
 			break;
 		    Add = True;
 		    break;
 		case NUMERICRANGE: {
 		    double 	tmpd;
 		    
-		    tmpd = strtod(Token, NULL);
+		    tmpd = Qddb_GetNum(Token);
 		    if (Arg->LowerExists == True && Arg->UpperExists == True) {
 			if (tmpd >= Arg->LowerNum && tmpd <= Arg->UpperNum)
 			    Add = True;
@@ -1336,15 +1408,13 @@ static KeyList *SecondarySearch(schema, NumberRead, ExistingList, Arg, attrnum)
                 Line = NULL; /* make 1st arg to strtok nil */
 	    }
 	    Free(inst);
-            Line = Buffer;
         }
-        fclose(FilePtr);
+	close(filedesc);
     }
     closedir(dp);
 
     /* Search through the additions directory.
      */
-    Line = Buffer;
     if ((dp = opendir(AddDir)) == NULL)
         PANIC("Cannot open Additions/ directory");
     while ((DirEntry = readdir(dp)) != NULL) {
@@ -1355,21 +1425,16 @@ static KeyList *SecondarySearch(schema, NumberRead, ExistingList, Arg, attrnum)
 		continue;
         sprintf(TmpDir, "%s%s", AddDir, dirnam);
 	AddCount = atoi(dirnam);
-        FilePtr = fopen(TmpDir, "r");
-        if (FilePtr == NULL)
-            PANIC("An existing file in Additions cannot be opened");
-	fgets(Line, bufsize, FilePtr); /* get %0 */
-        while (fgets(Line, bufsize, FilePtr) != NULL) {
+        filedesc = open(TmpDir, O_RDONLY);
+        if (filedesc < 0)
+            PANIC("An existing file in Changes/ cannot be opened");
+	entry = orig_entry;
+	Qddb_ReadEntry(filedesc, &entry, (off_t)0, (size_t)0, True);
+	orig_entry = entry++;
+        while (*entry != NULL) {
             char        *StringPtr;
 
-	    while (Line[strlen(Line)-2] == '\\' || Line[strlen(Line)-1] != '\n') {
-		if (strlen(Line) >= bufsize/2) {
-		    bufsize *= 2;
-		    Buffer = Line = (char *)Realloc(Buffer, bufsize);
-		}
-		if (fgets(Line+strlen(Line), bufsize-strlen(Line), FilePtr) == NULL)
-		    break;
-	    }
+	    Line = *entry++;
 	    StringPtr = Line;
 	    while (*StringPtr != '\0') {
 		if (_isupper(*StringPtr))
@@ -1405,20 +1470,38 @@ static KeyList *SecondarySearch(schema, NumberRead, ExistingList, Arg, attrnum)
 		    Add = True;
 		    break;
 	        case NUMERIC:
-		    if (strtod(Token, NULL) != Arg->Number)
+		    if (Qddb_GetNum(Token) != Arg->Number)
 			break;
 		    Add = True;
 		    break;
 		case NUMERICRANGE: {
 		    double 	tmpd;
 		    
-		    tmpd = strtod(Token, NULL);
-		    if (tmpd >= Arg->LowerNum && tmpd <= Arg->UpperNum)
+		    tmpd = Qddb_GetNum(Token);
+		    if (Arg->LowerExists == True && Arg->UpperExists == True) {
+			if (tmpd >= Arg->LowerNum && tmpd <= Arg->UpperNum)
+			    Add = True;
+		    } else if (Arg->LowerExists == True) {
+			if (tmpd >= Arg->LowerNum)
+			    Add = True;
+		    } else if (Arg->UpperExists == True) {
+			if (tmpd <= Arg->UpperNum)
+			    Add = True;
+		    } else 
 			Add = True;
 		    break;
 		}
 		case ALPHARANGE:
-		    if (strcmp(Arg->LowerStr, Token) <= 0 && strcmp(Arg->UpperStr, Token) >= 0)
+		    if (Arg->LowerExists == True && Arg->UpperExists == True) {
+			if (strcmp(Arg->LowerStr, Token) <= 0 && strcmp(Arg->UpperStr, Token) >= 0)
+			    Add = True;
+		    } else if (Arg->LowerExists == True) {
+			if (strcmp(Arg->LowerStr, Token) <= 0)
+			    Add = True;
+		    } else if (Arg->UpperExists == True) {
+			if (strcmp(Arg->UpperStr, Token) >= 0)
+			    Add = True;
+		    } else 
 			Add = True;
 		    break;
 		default:
@@ -1453,19 +1536,137 @@ static KeyList *SecondarySearch(schema, NumberRead, ExistingList, Arg, attrnum)
                 Line = NULL; /* make 1st arg to strtok nil */
             }
 	    Free(inst);
-            Line = Buffer;
         }
-        fclose(FilePtr);
+	close(filedesc);
     }
+    if (orig_entry != NULL)
+	Qddb_Free(QDDB_TYPE_ENTRY, orig_entry);
     closedir(dp);
     if (LastNode != NULL)
         LastNode->next = ExistingList;
     else
         NewList = ExistingList;
-    Free(Buffer);
+    NewList = Qddb_KeyListProcess(schema, NewList, NULL, QDDB_KEYLIST_PROC_SORT, 0);
     return NewList;
 }
 
+static KeyList *CachedSecondarySearch(schema, NumberRead, ExistingList, Arg, attrnum)
+    Schema		*schema;
+    size_t	    	*NumberRead;
+    KeyList        	*ExistingList;
+    Qddb_SearchArg	*Arg;
+    int			attrnum;
+{
+    KeyList		*tmplist, *lastnode;
+    char		*token;
+    int			i;
+
+#if defined(DEBUG)
+    printf("in CachedSecondarySearch\n");
+    fflush(stdout);
+#endif
+    Qddb_UpdateWordList(schema, 0, 0, NULL);
+    lastnode = NULL;
+    for (i = 0; i < schema->word_list_number_words; i++) {
+	Boolean		Add;
+
+	token = schema->word_list[i].word;
+	Add = False;
+	switch(Arg->Type) {
+#ifdef V8_REGEXP
+	case REGEXP_V8:
+	    /* Regexp or strcmp() */
+	    if (Arg->Reg == NULL ||
+		(regexec(Arg->Reg, token) != 1 ||
+		 token != Arg->Reg->startp[0] ||
+		 (token+strlen(token)) != Arg->Reg->endp[0]))
+		break;
+	    Add = True;
+	    break;
+#endif
+	case ALPHA: 
+	    if (strcmp(token, Arg->Alpha) != 0)
+		break;
+	    Add = True;
+	    break;
+	case NUMERIC:
+	    if (Qddb_GetNum(token) != Arg->Number)
+		break;
+	    Add = True;
+	    break;
+	case NUMERICRANGE: {
+	    double 	tmpd;
+	    
+	    tmpd = Qddb_GetNum(token);
+	    if (Arg->LowerExists == True && Arg->UpperExists == True) {
+		if (tmpd >= Arg->LowerNum && tmpd <= Arg->UpperNum)
+		    Add = True;
+	    } else if (Arg->LowerExists == True) {
+		if (tmpd >= Arg->LowerNum)
+		    Add = True;
+	    } else if (Arg->UpperExists == True) {
+		if (tmpd <= Arg->UpperNum)
+		    Add = True;
+	    } else 
+		Add = True;
+	    break;
+	}
+	case ALPHARANGE:
+	    if (Arg->LowerExists == True && Arg->UpperExists == True) {
+		if (strcmp(Arg->LowerStr, token) <= 0 && strcmp(Arg->UpperStr, token) >= 0)
+		    Add = True;
+	    } else if (Arg->LowerExists == True) {
+		if (strcmp(Arg->LowerStr, token) <= 0)
+		    Add = True;
+	    } else if (Arg->UpperExists == True) {
+		if (strcmp(Arg->UpperStr, token) >= 0)
+		    Add = True;
+	    } else 
+		Add = True;
+	    break;
+	default:
+	    PANIC("CachedSecondarySearch: bad Arg->Type");
+	}
+	if (Add == True && schema->word_list[i].list != NULL) {
+	    tmplist = Qddb_KeyListProcess(schema, schema->word_list[i].list, 
+					  NULL, QDDB_KEYLIST_PROC_NULLOP,
+					  QDDB_KEYLIST_FLAG_COPY);
+	    if (attrnum != -1) {
+		tmplist = Qddb_KeyListProcess(schema, tmplist, &attrnum,
+					      QDDB_KEYLIST_PROC_PRUNE,
+					      QDDB_KEYLIST_FLAG_PRUNEBYATTRINT);
+	    }	
+	    if (lastnode == NULL) {
+		if (ExistingList != NULL) {
+		    lastnode = ExistingList;
+		    while (lastnode->next != NULL) {
+			(*NumberRead)++;
+			lastnode = lastnode->next;
+		    }
+		    lastnode->next = tmplist;
+		} else {
+		    ExistingList = tmplist;
+		    if (tmplist != NULL) {
+			lastnode = tmplist;
+			while (lastnode->next != NULL) {
+			    (*NumberRead)++;
+			    lastnode = lastnode->next;
+			}
+		    }
+		}
+	    } else {
+		lastnode->next = tmplist;
+		while (lastnode->next != NULL) {
+		    (*NumberRead)++;
+		    lastnode = lastnode->next;
+		}
+	    }
+	}
+    }
+    ExistingList = Qddb_KeyListProcess(schema, ExistingList, NULL, QDDB_KEYLIST_PROC_SORT, 0);
+    return ExistingList;
+}
+
 static int RecurseSchema(Entries, start_entry, last_entry, List, start_list, last_list)
     SchemaNode			*Entries;
     int				start_entry, last_entry;
@@ -1479,7 +1680,7 @@ static int RecurseSchema(Entries, start_entry, last_entry, List, start_list, las
     for (i = start_entry; i <= last_entry; i++) {
 	if (Entries[i].Level == start_list && strcmp(Entries[i].Name, List[start_list]) == 0) {
 	    int			retval;
-	    if (start_list == last_list) /* && Entries[i].IsLeaf != False) */
+	    if (start_list == last_list)
 		return i;
 	    retval = RecurseSchema(Entries, i, last_entry, List, start_list+1, last_list);
 	    if (retval != -1)
