@@ -1,8 +1,7 @@
 
-
 /* qddb/Lib/LibQddb/Utils.c
  *
- * Copyright (C) 1993, 1994 Herrin Software Development, Inc.
+ * Copyright (C) 1993-1997 Herrin Software Development, Inc.
  * All rights reserved.
  *
  * This file is part of Qddb.
@@ -110,7 +109,6 @@ void Close(FileDesc)
     	perror("Close");
 }
 
-#ifdef QDDBDIRS
 /* Qddb_FindRelation --
  *	Return value must copied if routine reentered before the
  * string is used.
@@ -118,57 +116,65 @@ void Close(FileDesc)
 char *Qddb_FindRelation(RelationName)
     char		*RelationName;
 {
-    static char		NewName[MAXFILENAMELEN];
+    static char		NewName[MAXPATHLEN];
     static char		*QddbDirs = NULL; 
+    static char		*RealQddbDirs = NULL;
     char		*getenv();
-    char		*Dir, *TmpDirs;
+    char		*Dir, *TmpDirs, *last_ch, TmpRelationName[MAXPATHLEN], OldName[MAXPATHLEN];
 
-    if (RelationName[0] == '.') {
+    if (RelationName[0] == '\0')
+	return NULL;
+    strcpy(TmpRelationName, RelationName);
+    last_ch = TmpRelationName + strlen(TmpRelationName) - 1;
+    while (*last_ch == '/')
+	*last_ch-- = '\0';
+    last_ch = strrchr(TmpRelationName, (int)'/');
+    GETPWD(OldName);
+    if (chdir(TmpRelationName) == 0) {
 	GETPWD(NewName);
-	strcat(NewName, "/");
-	strcat(NewName, RelationName);
-	return NewName;
-    }
-    if (RelationName[0] == '/')
-    	return RelationName;
-    if (access(RelationName, 00) != -1) {
-	char		ch[MAXFILENAMELEN];
-
-	/* Check for normal directories with the same name as the
-	 * relation.  All relations must have a HashTable.
-	 */
-
-	strcpy(ch, RelationName); 
-	strcat(ch, "/HashTable");
-	if (access(ch, 00) != -1) {
-	    GETPWD(NewName);
-	    strcat(NewName, "/");
-	    strcat(NewName, RelationName);
+	if (access("HashTable", 00) != -1) {
+	    chdir(OldName);
 	    return NewName;
 	}
     }
-    if (QddbDirs == NULL) {
-    	QddbDirs = getenv("QDDBDIRS");
-    	if (QddbDirs == NULL) {
+    if (RealQddbDirs == NULL) {
+	RealQddbDirs = getenv("QDDBDIRS");
+	if (RealQddbDirs == NULL) {
 	    return NULL;
 	}
+	QddbDirs = Malloc(strlen(RealQddbDirs)+1);
     }
+    strcpy(QddbDirs, RealQddbDirs);
     TmpDirs = QddbDirs;
     Dir = strtok(TmpDirs, ":");
-    if (Dir == NULL)
-    	return RelationName;
+    if (Dir == NULL) {
+	strcpy(NewName, TmpRelationName);
+	if (chdir(NewName) == 0) {
+	    GETPWD(NewName);
+	    if (access("HashTable", 00) != -1) {
+		chdir(OldName);
+		return NewName;
+	    }
+	}
+    }
     TmpDirs = NULL;
     do {
-    	strcpy(NewName, Dir);
-    	strcat(NewName, "/");
-    	strcat(NewName, RelationName);
-    	if (access(NewName, 00) != -1)
-    		return NewName;
+	if (Dir[0] != '\0') {
+	    strcpy(NewName, Dir);
+	    strcat(NewName, "/");
+	    strcat(NewName, TmpRelationName);
+	    if (chdir(NewName) == 0) {
+		GETPWD(NewName);
+		if (access("HashTable", 00) != -1) {
+		    chdir(OldName);
+		    return NewName;
+		}
+	    }
+	}
     } while ((Dir=strtok(TmpDirs, ":")) != NULL);
+    chdir(OldName);
     return NULL;
 }
-#endif
-    
     	
 void Read(File, Buffer, SizeOfBuffer)
     int			File;
@@ -182,7 +188,7 @@ void Read(File, Buffer, SizeOfBuffer)
     	fprintf(stderr, "Attempted %d, actually read %d of filedesc %d\n", 
     		(int)SizeOfBuffer, NumberRead, File);
 	fflush(stderr);
-    	perror("Read():");
+    	perror("Read()");
     	exit(FILE_ERROR);
     }
 }
@@ -220,7 +226,6 @@ void PANIC(string)
     exit(2);
 }
 
-
 void Qddb_SplitBufferIntoLines(Lines, Buffer, BufLen)
     Entry		*Lines;
     char		*Buffer;
@@ -231,8 +236,8 @@ void Qddb_SplitBufferIntoLines(Lines, Buffer, BufLen)
     char		*ch;
     
     /* Split the buffer into lines.  A line is
-     * a set of characters beginning with '%' and
-     * ending in a newline.  Newlines embedded in
+     * a set of characters ending in a newline.
+     * Newlines embedded in
      * the line must be backslashed ('\').
      */
 
@@ -253,19 +258,20 @@ void Qddb_SplitBufferIntoLines(Lines, Buffer, BufLen)
     }
     LineTop++;
     *Lines = (Entry)Malloc(sizeof(char *)*(size_t)LineTop);
-    for (i = 0, j = 0, ch = Buffer; i < BufLen; i++)
+    for (i = 0, j = 0, ch = Buffer; i < BufLen; i++) {
 	if (Buffer[i] == '\0') { /* End of a line */
 	    size_t	len;
 
-	    if (i == BufLen - 1 && Buffer[i-1] == '\0')
+	    if (i == BufLen - 1 && Buffer[i-1] == '\0') {
 		break;
+	    }
 	    len = strlen(ch) + 1;
 	    (*Lines)[j] = Malloc(sizeof(char)*len);
-/* FIXME: need to check for reduced attribute identifiers here */
 	    strcpy((*Lines)[j], ch);
 	    ch = Buffer+i+1;
 	    j++;
 	}
+    }
 #if defined(DIAGNOSTIC)
     if (j >= LineTop)
 	PANIC("SplitBufferIntoLines(Utils.c): j >= LineTop\n");
@@ -281,22 +287,27 @@ void Qddb_ReadEntry(File, OldEntry, Start, Length, DoLocking)
     Boolean		DoLocking;
 {
     char		*Buffer;
+    struct stat		stat_buf;
 
 #if defined(DIAGNOSTIC)
     if (OldEntry == NULL)
 	PANIC("Qddb_ReadEntry: OldEntry == NULL\n");
 #endif
-#ifdef RECORDLOCKING
+#if defined(RECORDLOCKING)
     if (DoLocking == True)
 	LockSection(File, F_RDLCK, Start, (off_t)Length, True);
 #endif
-    if (Length == 0)
-	PANIC("Qddb_ReadEntry: trying to read a zero length file");
+    if (Length == 0) {
+	if (fstat(File, &stat_buf) != 0) {
+	    PANIC("Qddb_ReadEntry: cannot stat file\n");
+	}
+	Length = (size_t)stat_buf.st_size;
+    }
     Buffer = Malloc(sizeof(char)*(size_t)(Length+1));
     lseek(File, Start, 0);
     Read(File, Buffer, Length);
     Buffer[Length] = '\0';
-#ifdef RECORDLOCKING
+#if defined(RECORDLOCKING)
     if (DoLocking == True)
 	UnlockSection(File, Start, (off_t)Length);
 #endif
@@ -337,17 +348,19 @@ int Qddb_ReadEntryByKeyList(FileDesc, RelationFN, ThisEntry, TheKeyList, DoLocki
     	fd = OpenChangeFile(RelationFN, TheKeyList->Number, 0);
 	if (fd == -1)
 	    return -1;
-    	Qddb_ReadEntry(fd, ThisEntry, (off_t)0, SizeOfFile(fd), DoLocking);
+    	Qddb_ReadEntry(fd, ThisEntry, (off_t)0, (size_t)0, DoLocking);
     	Close(fd);
     	break;
     case ADDITION:
     	fd = OpenAdditionFile(RelationFN, TheKeyList->Number, 0);
 	if (fd == -1)
 	    return -1;
-    	Qddb_ReadEntry(fd, ThisEntry, (off_t)0, SizeOfFile(fd), DoLocking);
+    	Qddb_ReadEntry(fd, ThisEntry, (off_t)0, (size_t)0, DoLocking);
     	Close(fd);
     	break;
     default:
+	fprintf(stderr, "Qddb_ReadEntryByKeyList: type %d\n", QDDB_KEYLIST_TYPE(TheKeyList));
+	fflush(stderr);
     	PANIC("Qddb_ReadEntryByKeyList: BAD TYPE IN KeyList");
     }	
     return 0;
@@ -365,24 +378,25 @@ int Qddb_ReadEntryByKeyList2(FileDesc, RelationFN, ThisEntry, TheKeyList, DoLock
     switch(QDDB_KEYLIST_TYPE(TheKeyList)) {
     case ORIGINAL:
     	if (FileDesc == -1) {
-   		fd = OpenDatabase(RelationFN, 0);
-		if (fd == -1)
-		    return -1;
-    		Qddb_ReadEntry(fd, ThisEntry, TheKeyList->Start, 
-    			  TheKeyList->Length, DoLocking);
-    		Close(fd);
-    	} else 
-    		Qddb_ReadEntry(FileDesc, ThisEntry, TheKeyList->Start, 
-    			  TheKeyList->Length, DoLocking);
+	    fd = OpenDatabase(RelationFN, 0);
+	    if (fd == -1)
+		return -1;
+	    Qddb_ReadEntry(fd, ThisEntry, TheKeyList->Start, 
+			   TheKeyList->Length, DoLocking);
+	    Close(fd);
+    	} else {
+	    Qddb_ReadEntry(FileDesc, ThisEntry, TheKeyList->Start, 
+			   TheKeyList->Length, DoLocking);
+	}
 	if (EntryInvalid(*ThisEntry) == True) {
 	    return -1;
 	}
     	break;
     case CHANGE:
-    	Qddb_ReadEntry(FileDesc, ThisEntry, (off_t)0, SizeOfFile(FileDesc), DoLocking);
+    	Qddb_ReadEntry(FileDesc, ThisEntry, (off_t)0, (size_t)0, DoLocking);
     	break;
     case ADDITION:
-    	Qddb_ReadEntry(FileDesc, ThisEntry, (off_t)0, SizeOfFile(FileDesc), DoLocking);
+    	Qddb_ReadEntry(FileDesc, ThisEntry, (off_t)0, (size_t)0, DoLocking);
     	break;
     default:
     	PANIC("Qddb_ReadEntryByKeyList2: BAD TYPE IN KeyList");
@@ -400,14 +414,62 @@ void WriteEntry(schema, File, WhichEntry)
     Entry		WhichEntry;
 {
     Entry		AttributePtr = WhichEntry;
+    size_t		len = 0;
+    static char		*buffer = NULL;
+    static size_t	buffer_size = 0, content_size = 0;
+#define QDDB_BUFFER_SIZE (1024*1024)
 
-    Write(File, *AttributePtr, strlen(*AttributePtr));
-    Write(File, "\n", 1);
+    if (schema == NULL && qddb_stabilizing != 0) {
+	/* flush */
+	if (content_size > 0) {
+	    Write(File, buffer, content_size);
+	    content_size = 0;
+	    buffer[0] = '\0';
+	}
+	return;
+    }
+    if (buffer == NULL) {
+	buffer = Malloc(QDDB_BUFFER_SIZE);
+	buffer_size = QDDB_BUFFER_SIZE;
+	content_size = 0;
+	buffer[0] = '\0';
+    }
+    len = strlen(*AttributePtr) + 1;
     AttributePtr++;
     while (*AttributePtr != NULL) {
-	Write(File, *AttributePtr, strlen(*AttributePtr));
-    	Write(File, "\n", 1);
+	len += strlen(*AttributePtr) + 1;
     	AttributePtr++;
+    }    
+    while (len >= buffer_size) { /* increase buffer if necessary */
+	buffer_size *= 2;
+	buffer = Realloc(buffer, buffer_size);
+    }
+    if (qddb_stabilizing != 0) {
+	if (len > (buffer_size - content_size)) {
+	    Write(File, buffer, content_size);
+	    content_size = 0;
+	    buffer[0] = '\0';
+	}
+	AttributePtr = WhichEntry;
+	while (*AttributePtr != NULL) {
+	    strcat(buffer+content_size, *AttributePtr);
+	    content_size += strlen(*AttributePtr);
+	    strcat(buffer+content_size, "\n");
+	    content_size++;
+	    AttributePtr++;
+	}
+    } else { /* not stabilizing, don't buffer */
+	AttributePtr = WhichEntry;
+	while (*AttributePtr != NULL) {
+	    strcat(buffer+content_size, *AttributePtr);
+	    content_size += strlen(*AttributePtr);
+	    strcat(buffer+content_size, "\n");
+	    content_size++;
+	    AttributePtr++;
+	}
+	Write(File, buffer, content_size);
+	content_size = 0;
+	buffer[0] = '\0';
     }
 }
 
@@ -482,7 +544,7 @@ int StrcmpNoupper(string1, string2)
     return 0;
 }
 
-#ifdef RECORDLOCKING
+#if defined(RECORDLOCKING)
 int LockSection(File, LockType, Start, Length, Wait)
     int			File;
     int			LockType;
@@ -531,61 +593,26 @@ int UnlockSection(File, Start, Length)
 }
 #endif
 
-Boolean IsNumericString(string)
+Boolean Qddb_IsNumericString(string)
     char		*string;
 {
-    int			i, len;
-    char		*tmpstr;
-    Boolean		SawDot;
+    char			*endptr, *strend;
+    double			retval;
 
-    len = strlen(string);
-    tmpstr = string;
-    SawDot = False;
-    /* Optional negation */
-    if (*tmpstr == '-') {
-	i = 1;
-	tmpstr++;
+    retval = (double)strtol(string, &endptr, 0);
+    strend = string;
+    while (*strend != '\0') {
+	strend++;
+    }
+    if (endptr == strend) {
+	return True;
+    }
+    retval = strtod(string, &endptr);
+    if (endptr == strend) {
+	return True;
     } else {
-	i = 0;
+	return False;
     }
-    /* Check for integer */
-    while (*tmpstr != '\0') {
-	if (!isdigit(*tmpstr)) {
-	    if (isspace(*tmpstr))
-		return True;
-	    if (*tmpstr != '.') {
-		if (*tmpstr == 'e' || *tmpstr == 'E')
-		    break;
-		return False;
-	    } else {
-		if (SawDot == True)
-		    return False;
-		SawDot = True;
-	    }
-	}
-	tmpstr++;
-    }
-    /* Matched? */
-    if (*tmpstr == '\0')
-        return True;
-    /* Otherwise, we saw an 'e' or 'E' */
-    tmpstr++;
-    /* Optional sign or space */
-    while (*tmpstr != '\0' && isspace(*tmpstr))
-	tmpstr++;
-    if (*tmpstr == '-')
-	tmpstr++;
-    while (*tmpstr != '\0' && isspace(*tmpstr))
-	tmpstr++;
-    /* Integer portion of exponent */
-    while (*tmpstr != '\0') {
-	if (isspace(*tmpstr))
-	    return True;
-	if (!isdigit(*tmpstr))
-	    return False;
-	tmpstr++;
-    }
-    return True;
 }
 
 
@@ -601,12 +628,12 @@ void Qddb_InvalidateEntry(schema, Start, Length)
     off_t		Start;
     size_t		Length;
 {
-    char		*RelationFN = schema->RelationName, DatabaseFN[MAXFILENAMELEN];
+    char		*RelationFN = schema->RelationName, DatabaseFN[MAXPATHLEN];
     int			DatabaseFile = schema->database_fd;
     char		*Buffer;
     int			i;
 
-#ifdef RECORDLOCKING
+#if defined(RECORDLOCKING)
     LockSection(DatabaseFile, F_WRLCK, Start, (off_t)Length, True);
 #endif
     Buffer = Malloc(Length);
@@ -626,7 +653,7 @@ void Qddb_InvalidateEntry(schema, Start, Length)
     utime(DatabaseFN, NULL);
 #endif
 #endif
-#ifdef RECORDLOCKING
+#if defined(RECORDLOCKING)
     UnlockSection(DatabaseFile, Start, (off_t)Length);
 #endif
     Free(Buffer);
@@ -639,9 +666,9 @@ void Qddb_InvalidateEntryByNumber(schema, Number)
     size_t		Number;
 {
     char		*RelationFN = schema->RelationName;
-    char		DatabaseFN[MAXFILENAMELEN], *entry;
+    char		DatabaseFN[MAXPATHLEN], *entry;
     FILE		*DatabaseKeyFile;
-    int			i;
+    int			i, base;
     off_t		Start;
     size_t		Length;
     char		buf[BUFSIZ], *ptr = NULL;
@@ -654,8 +681,16 @@ void Qddb_InvalidateEntryByNumber(schema, Number)
 	i++;
     if (entry == NULL)
 	return;	/* must be an addition */
-    Start = (off_t)strtoul(buf, &ptr, 10);
-    Length = (size_t)strtoul(ptr, NULL, 10);
+    if (schema->UseCondensedIndexing == False) {
+	base = 10;
+    } else {
+	base = 62;
+    }
+    ptr = buf;
+    while (*ptr++ != ' ');
+    *ptr = '\0';
+    Start = Qddb_BaseToOffset(buf, base);
+    Length = Qddb_BaseToSize(ptr+1, base);
     Qddb_InvalidateEntry(schema, Start, Length);
     fclose(DatabaseKeyFile);
 }
@@ -664,13 +699,12 @@ void DeleteEntryByNumber(schema, Number)
     Schema		*schema;
     size_t		Number;
 {
-    char		DatabaseFN[MAXFILENAMELEN];
+    char		DatabaseFN[MAXPATHLEN];
     char		*RelationFN = schema->RelationName;
 
+    Qddb_UpdateWordList(schema, DELETION|CHANGE, (int)Number, NULL);
     sprintf(DatabaseFN, "%s/Changes/%d", RelationFN, (int)Number);
     unlink(DatabaseFN);
-    sprintf(DatabaseFN, "%s/Additions/%d", RelationFN, (int)Number);
-    unlink(DatabaseFN);
     Qddb_InvalidateEntryByNumber(schema, Number);
 }
 
@@ -678,38 +712,26 @@ void DeleteEntryByKeyList(schema, node)
     Schema		*schema;
     KeyList		*node;
 {
+    char		*RelationFN = schema->RelationName;
+    char		DatabaseFN[MAXPATHLEN];
+
     switch (QDDB_KEYLIST_TYPE(node)) {
     case ORIGINAL:
 	Qddb_InvalidateEntry(schema, node->Start, node->Length);
 	break;
     case CHANGE:
-    case ADDITION:
 	DeleteEntryByNumber(schema, node->Number);
 	break;
+    case ADDITION:
+	Qddb_UpdateWordList(schema, DELETION|ADDITION, (int)node->Number, NULL);
+	sprintf(DatabaseFN, "%s/Additions/%d", RelationFN, (int)node->Number);
+	unlink(DatabaseFN);
+	break;
     default:
 	;
     }
 }
 
-void StripBackslashes(buf)
-    char		*buf;
-{
-    char		*tmpbuf, *tmp, *tmpstart;
-
-    tmp = tmpstart = (char *)Malloc(strlen(buf)+1);
-    tmpbuf = buf;
-    while ((*tmp = *tmpbuf) != '\0') {
-	if (*tmpbuf == '\\' && *(tmpbuf+1) == '\n') {
-		tmpbuf++;
-	} else {
-	    tmp++;
-	    tmpbuf++;
-	}
-    }
-    strcpy(buf, tmpstart);
-    Free(tmpstart);
-}	    
-
 char **ParseKeyFile(fd)
     int			fd;
 {
@@ -737,10 +759,11 @@ char **ParseKeyFile(fd)
     return retval;
 }
 
-KeyList *CatRelation(RelationFN)
+KeyList *CatRelation(schema, RelationFN)
+    Schema              *schema;
     char		*RelationFN;
 {
-    char		KeyFN[BUFSIZ], **key_contents, buf[BUFSIZ];
+    char		KeyFN[BUFSIZ], **key_contents, **idx, buf[BUFSIZ];
     KeyList		*retval = NULL, *tmp, *last;
     int			fd;
     struct stat		stat_buf;
@@ -749,17 +772,26 @@ KeyList *CatRelation(RelationFN)
     struct dirent	*dir_ptr;
     DIR			*DIR_PTR;
     char		*ptr = NULL;
+    int                 base;
 
     strcpy(KeyFN, RelationFN);
     strcat(KeyFN,"/Database.key");
     if ((fd = Open(KeyFN, O_RDONLY, 0)) == -1)
 	return NULL;
-    key_contents = ParseKeyFile(fd);
+    idx = key_contents = ParseKeyFile(fd);
     number = 1;
     last = NULL;
-    while (*key_contents != NULL) {
-	start = (off_t)strtoul(*key_contents, &ptr, 10);
-	len = (size_t)strtoul(ptr, NULL, 10);
+    if (schema->UseCondensedIndexing == False) {
+	base = 10;
+    } else {
+	base = 62;
+    }
+    while (*idx != NULL) {
+	ptr = *idx;
+	while (*ptr++ != ' ');
+	*ptr = '\0';
+	start = Qddb_BaseToOffset(*idx, base);
+	len = Qddb_BaseToSize(ptr+1, base);
 	tmp = (KeyList *)Malloc(sizeof(KeyList));
 	tmp->Start = start;
 	tmp->Length = len;
@@ -777,7 +809,7 @@ KeyList *CatRelation(RelationFN)
 	    last->next = tmp;
 	    last = tmp;
 	}
-	key_contents++;
+	idx++;
     }
     sprintf(KeyFN, "%s/Additions", RelationFN);
     DIR_PTR = opendir(KeyFN);
@@ -836,7 +868,7 @@ int StringToInstance(str, Inst)
     size_t		Inst[];
 {
     char		*ch;
-    char		*buf = alloca(strlen(str)+1);
+    char		*buf = (char *)alloca(strlen(str)+1);
     int			i;
 
     strcpy(buf, str);
@@ -848,33 +880,58 @@ int StringToInstance(str, Inst)
     return i;
 }
 
+char *IntegerToString(integer)
+    int			integer;
+{
+    char		*retval, *ch;
+
+    retval = (char *)Malloc((size_t)MAXINTEGERLEN);
+    ch = FastIntegerToString(integer);
+    strcpy(retval, ch);
+    return retval;
+}
+
+char *FastIntegerToString(integer)
+    int			integer;
+{
+    static char		retval[MAXINTEGERLEN];
+    char		*forw, *back;
+    int			neg;
+
+    forw = retval;
+    if (integer < 0) {
+	*forw++ = '-';
+	integer = -integer;
+	neg = 1;
+    } else {
+	neg = 0;
+    }
+    while (integer > 0) {
+	*forw++ = '0' + (integer % 10);
+	integer /= 10;
+    }
+    *forw = '\0';
+    back = forw - 1;
+    forw = retval + neg;
+    while (forw < back) {
+	*forw ^= *back;
+	*back ^= *forw;
+	*forw++ ^= *back--;
+    }
+    return retval;
+}
+
 /* Rough date checking; don't actually check for a valid date */
 int Qddb_CheckDate(s)
     char			*s;
 {
-    char			buf[BUFSIZ];
     time_t			mytime, clockVal;
     long			zone;
 
     mytime = time(NULL);
     zone = Qddb_GetTimeZone(mytime);
-    if (strcmp(DEFAULT_DATE_FORMAT, "%d/%m/%y") == 0) {
-	int		mm, dd, yy;
-
-	if (sscanf(s, "%d/%d/%d", &dd, &mm, &yy) == 3) {
-	    sprintf(buf, "%d/%d/%d", mm, dd, yy);
-	    if (Qddb_GetDate(buf, mytime, zone, &clockVal) < 0) {
-		return 1;
-	    }
-	} else {
-	    if (Qddb_GetDate(s, mytime, zone, &clockVal) < 0) {
-		return 1;
-	    }
-	}
-    } else {
-	if (Qddb_GetDate(s, mytime, zone, &clockVal) < 0) {
-	    return 1;
-	}
+    if (Qddb_GetDate(s, mytime, zone, &clockVal) < 0) {
+	return 1;
     }
     return 0;
 }
@@ -894,34 +951,15 @@ char *Qddb_DateStringToTime(s)
     }	
     mytime = time(NULL);
     zone = Qddb_GetTimeZone(mytime);
-    if (strcmp(DEFAULT_DATE_FORMAT, "%d/%m/%y") == 0) {
-	int		mm, dd, yy;
-
-	if (sscanf(s, "%d/%d/%d", &dd, &mm, &yy) == 3) {
-	    sprintf(buf, "%d/%d/%d", mm, dd, yy);
-	    if (Qddb_GetDate(buf, mytime, zone, &clockVal) < 0) {
-		retval = Malloc(2);
-		strcpy(retval, "");
-		return retval;
-	    }
-	} else {
-	    if (Qddb_GetDate(s, mytime, zone, &clockVal) < 0) {
-		retval = Malloc(2);
-		strcpy(retval, "");
-		return retval;
-	    }
-	}
-    } else {
-	if (Qddb_GetDate(s, mytime, zone, &clockVal) < 0) {
-	    retval = Malloc(2);
-	    strcpy(retval, "");
-	    return retval;
-	}
+    if (Qddb_GetDate(s, mytime, zone, &clockVal) < 0) {
+	retval = Malloc(2);
+	strcpy(retval, "");
+	return retval;
     }
     if (sizeof(time_t) == sizeof(int))
-	sprintf(buf, "%u", (int)clockVal);
+	sprintf(buf, "%d", (int)clockVal);
     else if (sizeof(time_t) == sizeof(long))
-	sprintf(buf, "%lu", (long)clockVal);
+	sprintf(buf, "%ld", (long)clockVal);
     else
 	PANIC("sizeof(time_t) != sizeof(int) or sizeof(long)");
     retval = Malloc(sizeof(char)*(strlen(buf)+1));
@@ -934,6 +972,7 @@ char *Qddb_DateTimeToString(s, format)
     char 			*format;
 {
     char			buf[BUFSIZ], *retval;
+    int                         numchars;
     time_t			mytime;
     struct tm			*tm_buf;
 
@@ -943,10 +982,19 @@ char *Qddb_DateTimeToString(s, format)
 	return retval;
     }
     mytime = (time_t)strtoul(s, NULL, 10);
-    tm_buf = localtime(&mytime);
-    strftime(buf, BUFSIZ, format, tm_buf);
-    retval = Malloc(sizeof(char)*(strlen(buf)+1));
-    strcpy(retval, buf);
+    if (strcmp(format, "%s") == 0) {
+	sprintf(buf, "%d", (int)mytime);
+	numchars = strlen(buf);
+    } else {
+	tm_buf = localtime(&mytime);
+	numchars = strftime(buf, BUFSIZ, format, tm_buf);
+    }
+    retval = Malloc(sizeof(char)*(numchars+1));
+    if (numchars > 0) {
+	strcpy(retval, buf);
+    } else {
+	retval[0] = '\0';
+    }
     return retval;
 }
 
@@ -978,32 +1026,132 @@ int Qddb_TranslateExternalID(schema, id, attr, inst)
     size_t		*attr;
     char		**inst;
 {
-    char		*tok;
-    int			i, tmpattr;
-    size_t		attr_int[MAXIMUM_NUMBER_OF_SCHEMA_LEVELS];
-    size_t		inst_int[MAXIMUM_NUMBER_OF_SCHEMA_LEVELS];
+    char		*tok, inst_buf[BUFSIZ], attr_buf[BUFSIZ];
+    int			tmpattr;
+    size_t		inst_buflen, attr_buflen;
 
     if (*id == '%') {
 	id++;
     } else {
 	return -1;
     }
-    tok = strtok(id, ".");
-    i = 0;
-    while (tok != NULL) {
-	attr_int[i] = atoi(tok);
-	tok = strtok(NULL, ".");
-	inst_int[i++] = atoi(tok);
-	tok = strtok(NULL, ".");
+    inst_buf[0] = '\0';
+    inst_buflen = 0;
+    attr_buf[0] = '\0';
+    attr_buflen = 0;
+    tok = id;
+    while (*id != '\0' && *id != '.')
+	id++;
+    if (*id != '\0') {
+	*id = '\0';
+	id++;
     }
-    if ((tmpattr = Qddb_FindAttributeNumber(schema, attr_int, i)) == -1)
-	return -1;
-    if ((*inst = InstanceToString(inst_int, (size_t)i)) == NULL)
+    while (tok != NULL && *tok != '\0') {
+	strcpy(attr_buf+attr_buflen, tok);
+	attr_buflen += strlen(tok);
+	attr_buf[attr_buflen++] = '.';
+        attr_buf[attr_buflen] = '\0';
+	tok = id;
+	while (*id != '\0' && *id != '.')
+	    id++;
+	if (*id != '\0') {
+	    *id = '\0';
+	    id++;
+	}
+	strcpy(inst_buf+inst_buflen, tok);
+	inst_buflen += strlen(tok);
+	inst_buf[inst_buflen++] = '.';
+        inst_buf[inst_buflen] = '\0';
+	tok = id;
+	while (*id != '\0' && *id != '.')
+	    id++;
+	if (*id != '\0') {
+	    *id = '\0';
+	    id++;
+	}
+    }
+    attr_buflen--;
+    attr_buf[attr_buflen] = '\0';
+    if ((tmpattr = Qddb_FindAttributeNumberFromString(schema, attr_buf)) == -1)
 	return -1;
+    inst_buflen--;
+    inst_buf[inst_buflen] = '\0';
+    *inst = Malloc(inst_buflen+1);
+    strcpy(*inst, inst_buf);
     *attr = tmpattr;
     return 0;
 }
 
+
+int Qddb_FindAttributeNumber(schema, Attributes, HowMany)
+    Schema		*schema;
+    size_t		Attributes[];
+    int			HowMany;
+{
+    size_t		i, j;
+
+    for (i = 1; i <= schema->NumberOfAttributes; i++) {
+	for (j = 1; j < HowMany; j++) {
+	    if (schema->Entries[i].AncestorNumber[j] != Attributes[j-1])
+		break;
+	}
+	if (j == HowMany) {
+	    if (schema->Entries[i].Level != HowMany)
+		continue;
+	    if (Attributes[j-1] == schema->Entries[i].Number)
+		return i;
+	}
+    }
+    return -1;
+}
+
+int Qddb_FindAttributeNumberFromString(schema, str)
+    Schema			*schema;
+    char			*str;
+{
+#if defined(USE_TCL)
+    Tcl_HashEntry		*hash_entry;
+
+    hash_entry = Tcl_FindHashEntry(&(schema->TclHashTable), str);
+    if (hash_entry == NULL)
+	return -1;
+    return (int)Tcl_GetHashValue(hash_entry);
+#else
+    size_t			i;
+
+    for (i = 1; i <= schema->NumberOfAttributes; i++) {
+	if (strcmp(str, schema->Entries[i].Attribute) == 0) {
+	    return i;
+	}
+    }
+    return -1;
+#endif
+}
+
+int Qddb_GetInt(buf)
+    char			*buf;
+{
+    return (int)strtol(buf, NULL, 0);
+}
+
+double Qddb_GetNum(str)
+    char			*str;
+{
+    char			*endptr, *strend;
+    double			retval;
+
+    strend = str;
+    while (*strend != '\0') {
+	strend++;
+    }
+    retval = (double)strtol(str, &endptr, 0);
+    if (endptr == strend) {
+	return retval;
+    }
+    retval = strtod(str, NULL);
+    return retval;
+}
+
 char *Qddb_GetFile(fd)
     int				fd;
 {
@@ -1047,14 +1195,38 @@ size_t Qddb_ReadBuffer(buf, n)
     return i;
 }
 
-char *Qddb_GetBuffer(void)
+char *Qddb_GetBuffer _ANSI_ARGS_((void))
 {
     qddb_setbuf_buffer = Realloc(qddb_setbuf_buffer, qddb_setbuf_buflen+1);
     qddb_setbuf_bufsiz = qddb_setbuf_buflen+1;
     return qddb_setbuf_buffer;
 }
 
-void Qddb_InitBuffer(void)
+char *Qddb_GetBufferWithoutResize _ANSI_ARGS_((void))
+{
+    return qddb_setbuf_buffer;
+}
+
+size_t Qddb_GetBufferLength _ANSI_ARGS_((void))
+{
+    return qddb_setbuf_buflen;
+}
+
+void Qddb_SetBufferLength(len)
+    size_t			len;
+{
+    qddb_setbuf_buffer[len] = '\0';
+    qddb_setbuf_buflen = len;
+}
+
+void Qddb_SetBufferSize(len)
+    size_t			len;
+{
+    qddb_setbuf_bufsiz = len;
+}
+
+
+void Qddb_InitBuffer _ANSI_ARGS_((void))
 {
     qddb_setbuf_buffer = Malloc(BUFSIZ*sizeof(char));
     qddb_setbuf_buffer[0] = '\0';
@@ -1075,6 +1247,29 @@ void Qddb_ConcatBuffer(newbuf)
     qddb_setbuf_buflen += len;
 }
 
+void Qddb_ConcatPrintableBuffer(newbuf)
+    char			*newbuf;
+{
+    size_t			len = strlen(newbuf);
+    char			*ptr;
+
+    if (qddb_setbuf_buflen+len >= qddb_setbuf_bufsiz) {
+	qddb_setbuf_bufsiz += MAX(len+10, BUFSIZ);
+	qddb_setbuf_buffer = Realloc(qddb_setbuf_buffer, qddb_setbuf_bufsiz);
+    }
+    ptr = qddb_setbuf_buffer+qddb_setbuf_buflen;
+    while (*newbuf) {
+	if (isprint(*newbuf)) {
+	    *ptr++ = *newbuf;
+	} else {
+	    *ptr++ = ' ';
+	}
+	newbuf++;
+    }
+    *ptr = *newbuf;
+    qddb_setbuf_buflen += len;
+}
+
 char *Qddb_PrintErrno(qerrno) 
     int				qerrno;
 {
@@ -1088,7 +1283,6 @@ char *Qddb_PrintErrno(qerrno)
     }
 }
 
-
 char *Qddb_FindKey(string, len)
     char			*string;
     size_t			*len;
@@ -1100,7 +1294,10 @@ char *Qddb_FindKey(string, len)
     nlen = 0;
     for (; *string != '\0'; string++, nlen++) {
 	if (*string == '\\') {
-	    state = 1;
+	    if (state == 1)
+		state = 0;
+	    else
+		state = 1;
 	} else if (state == 1) {
 	    state = 0;
 	} else if (*string == ' ') {
@@ -1111,24 +1308,287 @@ char *Qddb_FindKey(string, len)
     return string;
 }
 
-char *Qddb_StripKey(string, len)
-    char			*string;
-    size_t			len;
+void Qddb_StripBackslashes(buf)
+    char		*buf;
 {
-    int				state;
-    char			*retval, *ch;
+    char		*tmp, *tmpbuf;
 
-    state = 0;
-    retval = Malloc(len+1);
-    for (ch = retval; *string != '\0'; string++, ch++) {
-	if (state == 0 && *string == '\\') {
-	    state = 1;
-	    ch--;
+    tmp = tmpbuf = buf;
+    while (*tmp != '\0') {
+	if (*tmp == '\\' && *(tmp+1) == '\n') {
+	    tmp++;
+	} else if (*tmp == '\\' && (*(tmp+1) == '\\' || *(tmp+1) == ' ')) {
+	    tmp++;
+	    *tmpbuf++ = *tmp++;
 	} else {
-	    state = 0;
-	    *ch = *string;
+	    *tmpbuf++ = *tmp++;
 	}
     }
-    *ch = '\0';
+    *tmpbuf = '\0';
+}	    
+
+char *Qddb_StripKey(buf, len)
+    char			*buf;
+    size_t			len;
+{
+    char		*tmp, *tmpbuf, *retval;
+
+    tmp = buf;
+    retval = tmpbuf = Malloc(len+1);
+    while (*tmp != '\0') {
+	if (*tmp == '\\' && *(tmp+1) == '\n') {
+	    tmp++;
+	} else if (*tmp == '\\' && (*(tmp+1) == '\\' || *(tmp+1) == ' ')) {
+	    tmp++;
+	    *tmpbuf++ = *tmp++;
+	} else {
+	    *tmpbuf++ = *tmp++;
+	}
+    }
+    *tmpbuf = '\0';
     return retval;
 }
+
+void Qddb_FillChars(buf, ch, len)
+    char			*buf;
+    int				ch, len;
+{
+    int				i;
+
+    for (i = 0; i < len; i++) {
+	buf[i] = (char)ch;
+    }
+}
+
+static char			*alphabet = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
+static int			indices[128] = {
+    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+    0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 0, 0, 0, 0, 0,
+    0, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50,
+    51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 0, 0, 0, 0, 0,
+    0, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24,
+    25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 0, 0, 0, 0, 0
+};
+
+char *Qddb_SizeToBase(num, base)
+    size_t                      num;
+    int                         base;
+{
+    static char                 retval[sizeof(off_t)*3], ch[sizeof(off_t)*3];
+    char                        *ch_ptr, *retval_ptr;
+    int                         i;
+
+    i = 0;
+    if (num == 0) {
+	retval[0] = '0';
+	retval[1] = '\0';
+	return retval;
+    }
+    ch_ptr = ch;
+    while (num > 0) {
+	*ch_ptr++ = alphabet[num % base];
+	num  /= base;
+	i++;
+    }
+    *ch_ptr = '\0';
+    retval_ptr = retval + i;
+    *retval_ptr-- = '\0';
+    ch_ptr = ch;
+    for (ch_ptr = ch; *ch_ptr != '\0'; ch_ptr++) {
+	*retval_ptr-- = *ch_ptr;
+    }
+    return retval;
+}
+
+
+size_t Qddb_BaseToSize(ch, base)
+    char                        *ch;
+    int                         base;
+{
+    size_t                      retval;
+    int				digit;
+
+    for (retval = 0; *ch != '\0' ; ch++) {
+	digit = indices[(int)(*ch)];
+	retval *= base;
+	retval += digit;
+    }
+    return retval;
+}
+
+char *Qddb_OffsetToBase(num, base)
+    off_t                       num;
+    int                         base;
+{
+    static char                 retval[sizeof(off_t)*3], ch[sizeof(off_t)*3];
+    char                        *ch_ptr, *retval_ptr;
+    int                         i;
+
+    i = 0;
+    if (num == 0) {
+	retval[0] = '0';
+	retval[1] = '\0';
+	return retval;
+    }
+    ch_ptr = ch;
+    while (num > 0) {
+	*ch_ptr++ = alphabet[num % base];
+	num  /= base;
+	i++;
+    }
+    *ch_ptr = '\0';
+    retval_ptr = retval + i;
+    *retval_ptr-- = '\0';
+    ch_ptr = ch;
+    for (ch_ptr = ch; *ch_ptr != '\0'; ch_ptr++) {
+	*retval_ptr-- = *ch_ptr;
+    }
+    return retval;
+}
+
+off_t Qddb_BaseToOffset(ch, base)
+    char                        *ch;
+    int                         base;
+{
+    off_t                      retval;
+    int                        digit;
+
+    for (retval = 0; *ch != '\0'; ch++) {
+	digit = indices[(int)(*ch)];
+	retval *= base;
+	retval += digit;
+    }
+    return retval;
+}
+
+Entry Qddb_ReadExcludeWords(schema, num)
+    Schema                     *schema;
+    size_t                     *num;
+{
+    size_t                     nmemb;
+    Entry                      idx, ExcludeEntry;
+    int                        ExcludeFile;
+    char                       ExcludeFN[MAXPATHLEN];
+
+    if (schema->UseExcludeWords == False)
+	return NULL;
+    *num = 0;
+    ExcludeEntry = NULL;
+    strcpy(ExcludeFN, schema->RelationName);
+    strcat(ExcludeFN, "/ExcludeWords");
+    if ((ExcludeFile = Open(ExcludeFN, O_RDONLY, 0)) != -1) {
+	Qddb_ReadEntry(ExcludeFile, &ExcludeEntry, (off_t)0, (size_t)0, True);
+	nmemb = 0;
+	idx = ExcludeEntry;
+	while (*idx != NULL) {
+	    char      *ch;
+
+	    ch = *idx;
+	    while (*ch != '\0') {
+		if (isupper(*ch))
+		    *ch = tolower(*ch);
+		ch++;
+	    }
+	    idx++;
+	    nmemb++;
+	}
+	Close(ExcludeFile);
+	*num = nmemb;
+	return ExcludeEntry;
+    }
+    return NULL;
+}
+
+
+/* Cache routines */
+#if defined(USE_TCL)
+typedef struct centry {
+    Entry			e;
+    time_t			tm;
+} qddb_cache_entry;
+
+static Tcl_HashTable 		qddb_cache;
+static int 			qddb_cache_init = 0;
+
+static qddb_cache_entry *Qddb_AddEntryToCache(fn)
+    char			*fn;
+{
+    qddb_cache_entry		*c;
+    int				FileDesc;
+    struct stat			stat_buf;
+    Tcl_HashEntry		*hash_entry;
+    int				newPtr;
+
+    c = (qddb_cache_entry *)Malloc(sizeof(qddb_cache_entry));
+    c->e = NULL;
+    /* open it up, lock it down */
+    FileDesc = open(fn, O_RDONLY);
+    LockSection(FileDesc, F_RDLCK, (off_t)0, (off_t)0, True);
+    Qddb_ReadEntry(FileDesc, &(c->e), (off_t)0, (size_t)0, False);
+    if (fstat(FileDesc, &stat_buf) != 0) {
+	PANIC("Qddb_GetEntryFromCache: cannot stat file\n");
+    }
+    c->tm = stat_buf.st_mtime;
+    UnlockSection(FileDesc, (off_t)0, (off_t)0);
+    close(FileDesc);
+    if (qddb_cache_init == 0) {
+	qddb_cache_init = 1;
+	Tcl_InitHashTable(&qddb_cache, TCL_STRING_KEYS);
+    }
+    hash_entry = Tcl_CreateHashEntry(&qddb_cache, fn, &newPtr);
+    Tcl_SetHashValue(hash_entry, (ClientData)c);
+    return c;
+}
+
+static qddb_cache_entry *Qddb_LookupEntryInCache(fn)
+    char			*fn;
+{
+    Tcl_HashEntry		*hash_entry;
+
+    if (qddb_cache_init != 0) {
+	hash_entry = Tcl_FindHashEntry(&qddb_cache, fn);
+	if (hash_entry == NULL)
+	    return NULL;
+	return (qddb_cache_entry *)Tcl_GetHashValue(hash_entry);
+    }
+    return NULL;
+}
+
+Entry Qddb_GetEntryFromCache(fn)
+    char			*fn;
+{
+    struct stat			stat_buf;
+    char			buf[BUFSIZ];
+    qddb_cache_entry		*cache;
+    int				FileDesc;
+
+    if (stat(fn, &stat_buf) != 0) {
+	sprintf(buf, "%s: %s", "Cannot stat file: ", fn);
+	PANIC(buf);
+    }
+    /* Check cache for entry:
+     *	1. If it doesn't, exist, add and return it.
+     *  2. If it does exist, check the time.
+     */
+    if ((cache = Qddb_LookupEntryInCache(fn)) == NULL) {
+	/* Doesn't exist, add it */
+	cache = Qddb_AddEntryToCache(fn);
+    } else {
+	if (cache->tm < stat_buf.st_mtime) { /* cache entry is old */
+	    /* open it up, lock it down */
+	    FileDesc = open(fn, O_RDONLY);
+	    LockSection(FileDesc, F_RDLCK, (off_t)0, (off_t)0, True);
+	    Qddb_ReadEntry(FileDesc, &(cache->e), (off_t)0, (size_t)0, False);
+	    if (fstat(FileDesc, &stat_buf) != 0) {
+		PANIC("Qddb_GetEntryFromCache: cannot stat requested file\n");
+	    }
+	    cache->tm = stat_buf.st_mtime;
+	    UnlockSection(FileDesc, (off_t)0, (off_t)0);
+	    close(FileDesc);
+	}
+    }
+    return cache->e;
+}
+#endif
